// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Cryptography;

using Smdn.IO;

namespace Smdn.Formats.Mime.Formatting {
 internal static class Encoder {
    internal static MimeMessageBuilder Encode(MimeMessageBuilder builder)
    {
      var encodedBuilder = new MimeMessageBuilder(builder.Headers.ConvertAll<MimeHeaderBuilder>(Encode),
                                                  builder.SubParts.ConvertAll<MimeMessageBuilder>(Encode));

      encodedBuilder.Content = builder.Content;
      encodedBuilder.Format = builder.Format;
      encodedBuilder.Charset = builder.Charset;
      encodedBuilder.MimeType = builder.MimeType;
      encodedBuilder.Boundary = builder.Boundary;
      encodedBuilder.TransferEncoding = builder.TransferEncoding;
      encodedBuilder.Disposition = builder.Disposition;

      return encodedBuilder;
    }

    private static MimeHeaderBuilder Encode(MimeHeaderBuilder builder)
    {
      var format = builder.Format ?? MimeFormat.Unspecified;

      return new MimeHeaderBuilder(builder.Name,
                                   EncodeHeaderFieldBody(builder, format.Folding - 1/*HT*/, builder.Name.Length + 2, format.GetHeaderFoldingString()));
    }

    internal static string Encode(MimeHeaderFragment fragment, MimeFormat format)
    {
      format = format ?? MimeFormat.Unspecified;

      return EncodeHeaderFieldBody(fragment, format.Folding - 1/*HT*/, 0, format.GetHeaderFoldingString());
    }

    private static string EncodeHeaderFieldBody(IHeaderFieldBody body, int foldingLimit, int foldingOffset, string foldingString)
    {
      switch (body.Encoding) {
        case MimeEncodingMethod.None:
          if (body.Charset == null)
            return body.Value;
          else
            return Charsets.ISO8859_1.GetString(body.Charset.GetBytes(body.Value));

        case MimeEncodingMethod.Base64:
        case MimeEncodingMethod.QuotedPrintable: {
          if (0 < foldingLimit) {
            if (body.Charset == null)
              return MimeEncoding.Encode(body.Value, body.Encoding, foldingLimit, foldingOffset, foldingString);
            else
              return MimeEncoding.Encode(body.Value, body.Encoding, body.Charset, foldingLimit, foldingOffset, foldingString);
          }
          else {
            if (body.Charset == null)
              return MimeEncoding.Encode(body.Value, body.Encoding);
            else
              return MimeEncoding.Encode(body.Value, body.Encoding, body.Charset);
          }
        }

        default:
          throw ExceptionUtils.CreateNotSupportedEnumValue(body.Encoding);
      }
    }

    internal static Stream CreateContentWritingStream(MimeMessageBuilder builder)
    {
      var format = (builder.Format ?? MimeFormat.Unspecified);

      if (builder.Content != null)
        builder.Content.Close();

      builder.Content = new ChunkedMemoryStream();

      return ContentEncodingStreamUtils.CreateEncodingStream(builder.Content, builder.TransferEncoding, format);
    }

    internal static StreamWriter CreateContentTextWriter(MimeMessageBuilder builder, bool leaveStreamOpen)
    {
      if (builder.TransferEncoding == ContentTransferEncodingMethod.Binary)
        throw new InvalidOperationException("can't create TextWriter from message of binary transfer encoding");

      var stream = CreateContentWritingStream(builder);

#if NET_4_5
      var writer = new StreamWriter(stream,
                                    builder.Charset ?? Charsets.ISO8859_1,
                                    1024,
                                    leaveStreamOpen);
#else
      if (leaveStreamOpen)
        stream = new NonClosingStream(stream);

      var writer = new StreamWriter(stream,
                                    builder.Charset ?? Charsets.ISO8859_1);
#endif

      writer.NewLine = (builder.Format ?? MimeFormat.Unspecified).GetEOLString();

      return writer;
    }

    internal static BinaryWriter CreateContentBinaryWriter(MimeMessageBuilder builder, bool leaveStreamOpen)
    {
      var stream = CreateContentWritingStream(builder);

#if NET_4_5
      return new BinaryWriter(stream, builder.Charset ?? Charsets.Latin1, leaveStreamOpen);
#else
      if (leaveStreamOpen)
        stream = new NonClosingStream(stream);

      return new BinaryWriter(stream, builder.Charset ?? Charsets.Latin1);
#endif
    }
  }
}