//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;

using Smdn.IO;

namespace Smdn.Formats.Mime.Formatting {
  public class MimeMessageBuilder : IDisposable {
    public MimeFormat Format {
      get; set;
    }

    public List<MimeHeaderBuilder> Headers {
      get { CheckDisposed(); return headers; }
    }

    internal Stream Content {
      get { CheckDisposed(); return content; }
      set { content = value; }
    }

    public MimeType MimeType {
      get; set;
    }

    public Encoding Charset {
      get; set;
    }

    public string Boundary {
      get; set;
    }

    public ContentTransferEncodingMethod TransferEncoding {
      get; set;
    }

    public MimeMessageDisposition Disposition {
      get; set;
    }

    public List<MimeMessageBuilder> SubParts {
      get { CheckDisposed(); return subParts; }
    }

    public MimeMessageBuilder()
      : this(Enumerable.Empty<MimeHeaderBuilder>(), Enumerable.Empty<MimeMessageBuilder>())
    {
      this.MimeType = MimeType.TextPlain;
    }

    public MimeMessageBuilder(IEnumerable<MimeHeaderBuilder> headers)
      : this(headers, Enumerable.Empty<MimeMessageBuilder>())
    {
      this.MimeType = MimeType.TextPlain;
    }

    public MimeMessageBuilder(IEnumerable<MimeMessageBuilder> subParts)
      : this(Enumerable.Empty<MimeHeaderBuilder>(), subParts)
    {
    }

    public MimeMessageBuilder(IEnumerable<MimeHeaderBuilder> headers,
                              IEnumerable<MimeMessageBuilder> subParts)
    {
      if (headers == null)
        throw new ArgumentNullException("headers");
      if (subParts == null)
        throw new ArgumentNullException("subParts");

      this.headers = new List<MimeHeaderBuilder>(headers);
      this.subParts = new List<MimeMessageBuilder>(subParts);

      this.Format = MimeFormat.Standard;
      this.MimeType = MimeType.MultipartMixed;
      this.Charset = Charsets.ISO8859_1;
      this.Boundary = null;
      this.TransferEncoding = ContentTransferEncodingMethod.SevenBit;
      this.Disposition = MimeMessageDisposition.None;
    }

    public void Dispose()
    {
      if (subParts != null) {
        foreach (var subPart in subParts) {
          subPart.Dispose();
        }

        subParts = null;
      }

      if (content != null) {
        content.Dispose();
        content = null;
      }

      disposed = true;
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

#region "Create*"
    public static MimeMessageBuilder CreateMultipart(IEnumerable<MimeMessageBuilder> subParts)
    {
      var builder = new MimeMessageBuilder(subParts);

      builder.Boundary = string.Format("----------------{0:D16}", DateTime.Now.Ticks);
      builder.MimeType = MimeType.MultipartMixed;

      builder.WriteContent("This is a multi-part message in MIME format.",
                           ContentTransferEncodingMethod.SevenBit,
                           Encoding.ASCII);

      return builder;
    }
    
    public static MimeMessageBuilder CreateAttachment(string file)
    {
      return CreateAttachment(file, null, false);
    }

    public static MimeMessageBuilder CreateAttachment(string file,
                                                      bool inline)
    {
      return CreateAttachment(file, null, inline);
    }

    public static MimeMessageBuilder CreateAttachment(string file,
                                                      string description,
                                                      bool inline)
    {
      using (var stream = File.OpenRead(file)) {
        return CreateAttachment(stream,
                                MimeType.FindMimeTypeByExtension(file),
                                Path.GetFileName(file),
                                description,
                                inline);
      }
    }

    public static MimeMessageBuilder CreateAttachment(Stream contentStream,
                                                      MimeType mimeType)
    {
      return CreateAttachment(contentStream, mimeType, null, null, false);
    }

    public static MimeMessageBuilder CreateAttachment(Stream contentStream,
                                                      MimeType mimeType,
                                                      bool inline)
    {
      return CreateAttachment(contentStream, mimeType, null, null, inline);
    }

    public static MimeMessageBuilder CreateAttachment(Stream contentStream,
                                                      MimeType mimeType,
                                                      string filename,
                                                      bool inline)
    {
      return CreateAttachment(contentStream, mimeType, filename, null, inline);
    }

    public static MimeMessageBuilder CreateAttachment(Stream contentStream,
                                                      MimeType mimeType,
                                                      string filename,
                                                      string description,
                                                      bool inline)
    {
      if (contentStream == null)
        throw new ArgumentNullException("contentStream");
      if (mimeType == null)
        throw new ArgumentNullException("mimeType");

      var disposition = inline ? MimeMessageDisposition.Inline : MimeMessageDisposition.Attachment;

      // TODO: Content-Type + name parameter
      var message = new MimeMessageBuilder(new[] {
        MimeHeaderBuilder.CreateContentType(mimeType, null, null, filename),
        MimeHeaderBuilder.CreateContentTransferEncoding(ContentTransferEncodingMethod.Base64),
        MimeHeaderBuilder.CreateContentDisposition(disposition, filename),
        new MimeHeaderBuilder(MimeHeaderNames.ContentDescription, description, MimeEncodingMethod.Base64, Charsets.UTF8),
      });

      message.MimeType = mimeType;
      message.Disposition = disposition;

      message.WriteContent(ContentTransferEncodingMethod.Base64, delegate(BinaryWriter writer) {
        contentStream.CopyTo(writer);
      });

      return message;
    }
#endregion

    public void Save(string file)
    {
      CheckDisposed();

      using (var stream = new FileStream(file, FileMode.Create, FileAccess.Write)) {
        Save(stream);
      }
    }
    
    public void Save(Stream stream)
    {
      CheckDisposed();

      if (stream == null)
        throw new ArgumentNullException("stream");

      Formatter.Format(Encoder.Encode(this), stream);
    }

    /// <remarks>do not close supplied writer</remarks>
    public void WriteContent(ContentTransferEncodingMethod transferEncoding,
                             Encoding charset,
                             Action<StreamWriter> write)
    {
      CheckDisposed();

      if (charset == null)
        throw new ArgumentNullException("charset");
      if (write == null)
        throw new ArgumentNullException("write");

      this.Charset = charset;
      this.TransferEncoding = transferEncoding;

      using (var writer = Encoder.CreateContentTextWriter(this, true)) {
        var baseStream =
#if NET_4_5
          writer.BaseStream as ContentEncodingStream;
#else
        (writer.BaseStream as NonClosingStream).InnerStream as ContentEncodingStream;
#endif

        write(writer);

        writer.Close();

        if (baseStream != null)
          baseStream.FlushFinalBlock();
      }

    }

    /// <remarks>do not close supplied writer</remarks>
    public void WriteContent(ContentTransferEncodingMethod transferEncoding,
                             Action<BinaryWriter> write)
    {
      CheckDisposed();

      if (write == null)
        throw new ArgumentNullException("write");

      this.Charset = Charsets.ISO8859_1;
      this.TransferEncoding = transferEncoding;

      using (var writer = Encoder.CreateContentBinaryWriter(this, true)) {
        var baseStream =
#if NET_4_5
          writer.BaseStream as ContentEncodingStream;
#else
          (writer.BaseStream as NonClosingStream).InnerStream as ContentEncodingStream;
#endif

        write(writer);

        writer.Close();

        if (baseStream != null)
          baseStream.FlushFinalBlock();
      }
    }

    public void WriteContent(string content,
                             ContentTransferEncodingMethod transferEncoding,
                             Encoding charset)
    {
      WriteContent(transferEncoding, charset, delegate(StreamWriter writer) {
        writer.Write(content);
      });
    }

    public void WriteContent(byte[] content,
                             ContentTransferEncodingMethod transferEncoding)
    {
      WriteContent(content, 0, content.Length, transferEncoding);
    }

    public void WriteContent(byte[] content,
                             int index,
                             int count,
                             ContentTransferEncodingMethod transferEncoding)
    {
      WriteContent(transferEncoding, delegate(BinaryWriter writer) {
        writer.Write(content, index, count);
      });
    }

    public IEnumerable<MimeHeaderBuilder> GetHeaders(string name)
    {
      CheckDisposed();

      foreach (var header in headers) {
        if (header.IsNameEquals(name))
          yield return header;
      }
    }

    public override string ToString()
    {
      CheckDisposed();

      using (var stream = new ChunkedMemoryStream()) {
        Save(stream);

        stream.Position = 0L;

        return (new StreamReader(stream, Charsets.Latin1)).ReadToEnd();
      }
    }

    private bool disposed;
    private /*readonly*/ List<MimeHeaderBuilder> headers;
    private /*readonly*/ Stream content;
    private /*readonly*/ List<MimeMessageBuilder> subParts;
  }
}