//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.ObjectModel;
using System.IO;

using Smdn.IO;

namespace Smdn.Formats.Mime {
  public class AttachedFile : MailBase, IMimeMessage {
    /*
     * class members
     */
    public static AttachedFile CreateFrom(MimeMessage baseMessage)
    {
      return new AttachedFile(null, baseMessage);
    }

    /*
     * instance members
     */
    public Mail Mail {
      get { return mail; }
    }

    public MimeHeaderCollection Headers {
      get { return BaseMessage.Headers; }
    }

    public MimeType MimeType {
      get { return BaseMessage.MimeType; }
    }

    public MimeMessageDisposition Disposition {
      get { return BaseMessage.Disposition; }
    }

    ReadOnlyCollection<MimeMessage> IMimeMessage.SubParts {
      get { return BaseMessage.SubParts; }
    }

    MimeMessage IMimeMessage.ParentPart {
      get { return BaseMessage.RootPart; }
    }

    MimeMessage IMimeMessage.RootPart {
      get { return BaseMessage.RootPart; }
    }

    MimeMessage IMimeMessage.EncapsulatedMessage {
      get { return BaseMessage.EncapsulatedMessage; }
    }

    public string FileName {
      get {
        // Content-Disposition: xxx; filename=
        var filename = contentDisposition.GetParameter("filename");

        if (filename != null)
          return filename;

        // Content-Type: xxx; name=
        var name = contentType.GetParameter("name");

        if (name != null)
          return name;

        return null;
      }
    }

    public DateTimeOffset? CreationDate {
      get { return ParseDateParameter("creation-date"); }
    }

    public DateTimeOffset? ModificationDate {
      get { return ParseDateParameter("modification-date"); }
    }

    public DateTimeOffset? ReadDate {
      get { return ParseDateParameter("read-date"); }
    }

    public long? Size {
      get {
        var size = contentDisposition.GetParameter("size");

        if (string.IsNullOrEmpty(size))
          return null;

        long ret;

        if (long.TryParse(size, out ret))
          return ret;
        else
          return null;
      }
    }

    internal AttachedFile(Mail mail, MimeMessage baseMessage)
      : base(baseMessage)
    {
      this.mail = mail;

      if (!baseMessage.Headers.TryGetHeader(MimeHeaderNames.ContentDisposition, out contentDisposition))
        contentDisposition = MimeHeader.CreateEmpty(MimeHeaderNames.ContentDisposition);

      if (!baseMessage.Headers.TryGetHeader(MimeHeaderNames.ContentType, out contentType))
        contentType = MimeHeader.CreateEmpty(MimeHeaderNames.ContentType);
    }

    public FileInfo Save()
    {
      CheckDisposed();

      var file = new FileInfo(GetDefaultFileName());

      SaveAs(file);

      return file;
    }

    public FileInfo Save(string saveToDirectory)
    {
      if (saveToDirectory == null)
        throw new ArgumentNullException("saveToDirectory");
      if (saveToDirectory.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("saveToDirectory");

      CheckDisposed();

      var file = new FileInfo(Path.Combine(saveToDirectory, GetDefaultFileName()));

      file.Directory.Create();

      SaveAs(file);

      return file;
    }

    internal void SaveAs(FileInfo file)
    {
      if (BaseMessage.EncapsulatedMessage != null)
        // message/*
        BaseMessage.EncapsulatedMessage.Save(file);
      else
        SaveContent(file, true);
    }

    internal string GetDefaultFileName()
    {
      /*
       * filename priority
       * 
       * 1. Content-Disposition, filename param
       * 2. Content-Type, name param
       * 3. Subject + part specifier
       * 4. Message-ID + part specifier
       * 5. part specifier
       */

      var filename = FileName;

      if (filename != null)
        return Path.GetFileName(filename); // remove directory name

      if (mail == null) {
        filename = string.Empty;
      }
      else {
        if (string.IsNullOrEmpty(filename = mail.Subject))
          if (string.IsNullOrEmpty(filename = mail.MessageId))
            filename = string.Empty;
      }

      if (0 < filename.Length)
        filename += ".";

      filename += BaseMessage.PartSpecifier;

      return string.Concat(PathUtils.RemoveInvalidFileNameChars(filename),
                           MimeTypeUtils.FindExtensionFromMimeType(BaseMessage.MimeType));
    }

    private DateTimeOffset? ParseDateParameter(string parameterName)
    {
      try {
        return DateTimeConvert.FromRFC822DateTimeOffsetStringNullable(contentDisposition.GetParameter(parameterName));
      }
      catch (FormatException) {
        return null;
      }
    }

    private readonly Mail mail;
    private readonly MimeHeader contentDisposition;
    private readonly MimeHeader contentType;
  }
}

