// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Text;

using Smdn.IO;

namespace Smdn.Formats.Mime {
  public class MimeMessage : IMimeMessage, IDisposable {
    internal static readonly Encoding DefaultCharset = Charsets.ISO8859_1;
    internal static readonly MimeType DefaultMimeType = MimeType.TextPlain;

    public MimeHeaderCollection Headers {
      get { CheckDisposed(); return headers; }
    }

    internal Stream Content {
      get { CheckDisposed(); return content; }
      set { content = value; }
    }

    public ReadOnlyCollection<MimeMessage> SubParts {
      get { CheckDisposed(); return subParts; }
    }

    public MimeMessage ParentPart {
      get { CheckDisposed(); return parentPart; }
      internal set { parentPart = value; }
    }

    public MimeMessage RootPart {
      get {
        CheckDisposed();

        var root = this;

        for (;;) {
          var p = root.ParentPart;

          if (p == null)
            return root;
          else
            root = p;
        }
      }
    }

    public MimeMessage[] AllParts {
      get { return GetFlattenSubParts().ToArray(); }
    }

    public string PartSpecifier {
      get {
        CheckDisposed();

        if (parentPart == null) {
          return string.Empty;
        }
        else {
          var ret = parentPart.PartSpecifier;

          if (0 < ret.Length)
            ret += ".";

          return ret + (parentPart.subParts.IndexOf(this) + 1).ToString();
        }
      }
    }

    public MimeMessage EncapsulatedMessage {
      get { CheckDisposed(); return encapsulatedMessage; }
    }

    public MimeType MimeType {
      get { return mimeType; }
    }

    public Encoding Charset {
      get { return charset; }
    }

    public string Boundary {
      get { return boundary; }
    }

    public ContentTransferEncodingMethod TransferEncoding {
      get { return transferEncoding; }
    }

    public MimeMessageDisposition Disposition {
      get { return disposition; }
    }

    internal MimeMessage(MimeHeaderCollection headers)
    {
      this.headers = headers;
      this.content = null;

      SetSubParts(null);
    }

    internal void SetSubParts(IEnumerable<MimeMessage> subParts)
    {
      if (subParts == null)
        this.subParts = (new List<MimeMessage>(1)).AsReadOnly(); // empty list
      else
        this.subParts = subParts.ToList().AsReadOnly();
    }

    internal void SetEncapsulatedMessage(MimeMessage encapsulatedMessage)
    {
      this.encapsulatedMessage = encapsulatedMessage;
    }

    internal void SetContentType(MimeType mimeType, Encoding charset, string boundary)
    {
      this.mimeType = mimeType ?? DefaultMimeType;
      this.charset = charset ?? DefaultCharset;
      this.boundary = boundary;
    }

    internal void SetContentTransferEncoding(ContentTransferEncodingMethod transferEncoding)
    {
      this.transferEncoding = transferEncoding;
    }

    internal void SetContentDisposition(MimeMessageDisposition disposition)
    {
      this.disposition = disposition;
    }

    internal IEnumerable<MimeMessage> GetFlattenSubParts()
    {
      CheckDisposed();

      var stack = new Stack<MimeMessage>();

      for (var p = this;;) {
        for (var index = p.subParts.Count - 1; 0 <= index; index--) {
          stack.Push(p.subParts[index]);
        }

        if (stack.Count == 0)
          yield break;

        p = stack.Pop();

        yield return p;
      }
    }

    public void Dispose()
    {
      if (subParts != null) {
        foreach (var subPart in subParts) {
          subPart.Dispose();
        }

        subParts = null;
      }

      if (encapsulatedMessage != null) {
        encapsulatedMessage.Dispose();
        encapsulatedMessage = null;
      }

      if (content != null) {
        if (this == RootPart) {
          var rootContentStream = content as PartialStream;

          if (rootContentStream != null)
            // InnerStream must be FileStream
            rootContentStream.InnerStream.Close();
        }

        content.Close();
        content = null;
      }

      disposed = true;
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private const bool loadContentToMemoryParameterDefault = false;

    public static MimeMessage Load(string file)
    {
      return Load(file, loadContentToMemoryParameterDefault, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage Load(string file, bool loadContentToMemory)
    {
      return Load(file, loadContentToMemory, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage Load(string file,
                                   Encoding fallbackCharset)
    {
      return Load(file, loadContentToMemoryParameterDefault, fallbackCharset);
    }

    public static MimeMessage Load(string file,
                                   bool loadContentToMemory,
                                   Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return Load(file, delegate(string charsetName) {
        return fallbackCharset;
      }, null);
    }

    public static MimeMessage Load(string file,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      return Load(file, loadContentToMemoryParameterDefault, selectFallbackCharset, null);
    }

    public static MimeMessage Load(string file,
                                   bool loadContentToMemory,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      return Load(file, loadContentToMemory, selectFallbackCharset, null);
    }

    public static MimeMessage Load(string file,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      return Load(file, loadContentToMemoryParameterDefault, selectFallbackCharset, decodeMalformedHeaderValue);
    }

    public static MimeMessage Load(string file,
                                   bool loadContentToMemory,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      FileStream stream = null;

      try {
        stream = File.OpenRead(file);

        return LoadCore(stream, loadContentToMemory, selectFallbackCharset, decodeMalformedHeaderValue);
      }
      catch {
        if (stream != null) {
          stream.Close();
          stream = null;
        }

        throw;
      }
      finally {
        if (loadContentToMemory && stream != null) {
          stream.Close();
          stream = null;
        }
      }
    }

    public static MimeMessage LoadMessage(string message)
    {
      return LoadMessage(message, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage LoadMessage(string message,
                                          Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return LoadMessage(message, delegate(string charsetName) {
        return fallbackCharset;
      }, null);
    }

    public static MimeMessage LoadMessage(string message,
                                          EncodingSelectionCallback selectFallbackCharset)
    {
      return LoadMessage(message, selectFallbackCharset, null);
    }

    public static MimeMessage LoadMessage(string message,
                                          EncodingSelectionCallback selectFallbackCharset,
                                          MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      const int lengthThreshold = 20480;

      if (message == null)
        throw new ArgumentNullException("message");

      Stream messageStream = null;

      try {
        if (message.Length <= lengthThreshold) {
          messageStream = new MemoryStream(Charsets.ISO8859_1.GetBytes(message));
        }
        else {
          messageStream = new ChunkedMemoryStream();

          var writer = new StreamWriter(messageStream, Charsets.ISO8859_1);

          writer.Write(message);
          writer.Flush();
          // writer.Close(); // leave stream open

          messageStream.Position = 0L;
        }

        return LoadCore(messageStream,
                        true,
                        selectFallbackCharset,
                        decodeMalformedHeaderValue);
      }
      finally {
        if (messageStream != null)
          messageStream.Close();
      }
    }

    public static MimeMessage Load(Stream stream)
    {
      return LoadCore(stream, true, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage Load(Stream stream,
                                   Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return LoadCore(stream, true, delegate(string charsetName) {
        return fallbackCharset;
      }, null);
    }

    public static MimeMessage Load(Stream stream,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      return LoadCore(stream, true, selectFallbackCharset, null);
    }

    public static MimeMessage Load(Stream stream,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      return LoadCore(stream, true, selectFallbackCharset, decodeMalformedHeaderValue);
    }

    private static MimeMessage LoadCore(Stream stream,
                                        bool loadContentToMemory,
                                        EncodingSelectionCallback selectFallbackCharset,
                                        MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      return Parser.Parse(new LooseLineOrientedStream(stream),
                          loadContentToMemory,
                          selectFallbackCharset,
                          decodeMalformedHeaderValue);
    }

    public static MimeMessage ReassembleFragmentedMessage(IMimeMessage fragmentedMessage1,
                                                          IMimeMessage fragmentedMessage2,
                                                          params IMimeMessage[] fragmentedMessages)
    {
      if (fragmentedMessage1 == null)
        throw new ArgumentNullException("fragmentedMessage1");
      if (fragmentedMessage2 == null)
        throw new ArgumentNullException("fragmentedMessage2");

      var fragments = new List<IMimeMessage>(2 + (fragmentedMessages == null ? 0 : fragmentedMessages.Length));

      fragments.Add(fragmentedMessage1);
      fragments.Add(fragmentedMessage2);

      if (fragmentedMessages != null && 0 < fragmentedMessages.Length)
        fragments.AddRange(fragmentedMessages);

      return ReassembleFragmentedMessage(fragments);
    }

    public static MimeMessage ReassembleFragmentedMessage(IEnumerable<IMimeMessage> fragmentedMessages)
    {
      if (fragmentedMessages == null)
        throw new ArgumentNullException("fragmentedMessages");

      return ReassembleFragmentedMessage(null, fragmentedMessages);
    }

    public MimeMessage ReassembleFragmentedMessage(IMimeMessage fragmentedMessage,
                                                   params IMimeMessage[] fragmentedMessages)
    {
      CheckDisposed();

      if (!this.mimeType.TypeEqualsIgnoreCase(MimeType.MessagePartial))
        throw new InvalidOperationException("instance is not fragmented message");
      if (fragmentedMessage == null)
        throw new ArgumentNullException("fragmentedMessage");

      var fragments = new List<IMimeMessage>(2 + (fragmentedMessages == null ? 0 : fragmentedMessages.Length));

      fragments.Add(this);
      fragments.Add(fragmentedMessage);

      if (fragmentedMessages != null && 0 < fragmentedMessages.Length)
        fragments.AddRange(fragmentedMessages);

      return ReassembleFragmentedMessage(GetPartialID(this), fragments);
    }

    private static string GetPartialID(IMimeMessage mimeMessage)
    {
      MimeHeader contentType;

      if (mimeMessage.Headers.TryGetHeader(MimeHeaderNames.ContentType, out contentType))
        return contentType.GetParameter("id");
      else
        return null;
    }

    private static int GetPartialParameter(IMimeMessage mimeMessage, string parameter)
    {
      MimeHeader contentType;

      if (mimeMessage.Headers.TryGetHeader(MimeHeaderNames.ContentType, out contentType)) {
        int number;

        if (int.TryParse(contentType.GetParameter(parameter), out number))
          return number;
      }

      return -1;
    }

    private static MimeMessage ReassembleFragmentedMessage(string id, IEnumerable<IMimeMessage> fragmentedMessages)
    {
      var orderedFragments = new SortedList<int, IMimeMessage>();
      int total = 0;

      foreach (var fragmentedMessage in fragmentedMessages) {
        if (!MimeType.MessagePartial.EqualsIgnoreCase(fragmentedMessage.MimeType))
          // is not 'message/partial'
          continue;

        if (id != null && !string.Equals(id, GetPartialID(fragmentedMessage), StringComparison.Ordinal))
          // id not specified or different
          continue;

        var number = GetPartialParameter(fragmentedMessage, "number");

        if (0 < number)
          orderedFragments[number] = fragmentedMessage;

        var t = GetPartialParameter(fragmentedMessage, "total");

        if (0 < t)
          total = t;
      }

      if (!orderedFragments.ContainsKey(1))
        throw new ArgumentException("fragmentedMessages does not contain enclosing message", "fragmentedMessages");

      /*
       * reassemble message
       */
      MimeMessage reassembledMessage = null;

      using (var reassembledMessageStream = new ChunkedMemoryStream()) {
        for (var number = 1;; number++) {
          if (!orderedFragments.ContainsKey(number))
            throw new ArgumentException(string.Format("missing fragment (number={0})", number), "fragmentedMessages");

          // reassemble partial content
          var message = orderedFragments[number] as MimeMessage;

          if (message == null) {
            orderedFragments[number].ReadContent(false, delegate(Stream stream) {
              stream.CopyTo(reassembledMessageStream);
            });
          }
          else {
            message.CopyContentTo(reassembledMessageStream, false);
          }

          if (0 < total) {
            if (total <= number)
              break;
          }
          else {
            if (orderedFragments.Count <= number)
              break;
          }
        }

        reassembledMessageStream.Position = 0L;

        reassembledMessage = MimeMessage.Load(reassembledMessageStream);
      }

      /*
       * reassemble header fields
       */
      var enclosingMessage = orderedFragments[1];
      var reassembledHeaders = new List<MimeHeader>();

      foreach (var header in enclosingMessage.Headers) {
        if (!FilterReassemblingMessageHeaders(header))
          reassembledHeaders.Add(header.Clone());
      }

      foreach (var header in reassembledMessage.Headers) {
        if (FilterReassemblingMessageHeaders(header))
          reassembledHeaders.Add(header.Clone());
      }

      reassembledMessage.headers = new MimeHeaderCollection(reassembledHeaders);

      return reassembledMessage;
    }

    private static bool FilterReassemblingMessageHeaders(MimeHeader header)
    {
      if (header.Name.StartsWith("Content-", StringComparison.OrdinalIgnoreCase))
        return true;
      else if (string.Equals("Subject", header.Name, StringComparison.OrdinalIgnoreCase))
        return true;
      else if (string.Equals("Message-ID", header.Name, StringComparison.OrdinalIgnoreCase))
        return true;
      else if (string.Equals("Encrypted", header.Name, StringComparison.OrdinalIgnoreCase))
        return true;
      else if (string.Equals("MIME-Version", header.Name, StringComparison.OrdinalIgnoreCase))
        return true;
      else
        return false;
    }

    private const bool readContentDecodeParameterDefault = true;

    public void ReadContent(Action<Stream> read)
    {
      ReadContent(readContentDecodeParameterDefault, read);
    }

    public void ReadContent(bool decode, Action<Stream> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentReadingStream(this, decode, true));
    }

    public void ReadContent(Action<StreamReader> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentTextReader(this));
    }

    public void ReadContent(Action<BinaryReader> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentBinaryReader(this));
    }

    public string ReadContentAsText()
    {
      CheckDisposed();

      return Decoder.CreateContentTextReader(this).ReadToEnd();
    }

    public byte[] ReadContentAsBinary()
    {
      CheckDisposed();

      var contentStream = Decoder.CreateContentReadingStream(this, true, false);
      var resultStream = new MemoryStream((int)contentStream.Length);

      contentStream.CopyTo(resultStream);

      resultStream.Close();

      return resultStream.ToArray();
    }

    public void CopyContentTo(Stream stream)
    {
      CopyContentTo(stream, readContentDecodeParameterDefault);
    }

    public void CopyContentTo(Stream stream, bool decode)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");

      Decoder.CreateContentReadingStream(this, decode, false).CopyTo(stream);
    }

    public void SaveContent(string path)
    {
      SaveContent(new FileInfo(path), readContentDecodeParameterDefault);
    }

    public void SaveContent(string path, bool decode)
    {
      SaveContent(new FileInfo(path), decode);
    }

    internal void SaveContent(FileInfo file, bool decode)
    {
      CheckDisposed();

      using (var fileStream = file.Open(FileMode.Create, FileAccess.Write)) {
        Decoder.CreateContentReadingStream(this, decode, false).CopyTo(fileStream);
      }
    }

    public void Save(string path)
    {
      Save(new FileInfo(path));
    }

    internal void Save(FileInfo file)
    {
      CheckDisposed();

      using (var fileStream = file.Open(FileMode.Create, FileAccess.Write)) {
        Save(fileStream);
      }
    }

    private static readonly byte[] crlf = Octets.GetCRLF();

    public void Save(Stream stream)
    {
      CheckDisposed();

      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      // write headers
      foreach (var header in headers) {
        stream.Write(header.RawHeaderField.Segment);
      }

      stream.Write(crlf, 0, crlf.Length);

      // write content or preamble
      CopyContentTo(stream, false);

      // write enclosed messages
      if (0 < subParts.Count) {
        if (0 < content.Length)
          stream.Write(crlf, 0, crlf.Length);

        var delimiter = "--" + boundary;
        var delimiterLine = ByteString.CreateImmutable(delimiter + Chars.CRLF);

        foreach (var subPart in subParts) {
          stream.Write(delimiterLine.Segment);

          subPart.Save(stream);

          stream.Write(crlf, 0, crlf.Length);
        }

        // close delimiter
        stream.Write(ByteString.CreateImmutable(delimiter + "--" + Chars.CRLF).Segment);

        // TODO: write epilogue
      }
      else if (encapsulatedMessage != null) {
        encapsulatedMessage.Save(stream);
      }

      stream.Flush();
    }

    public override string ToString()
    {
      CheckDisposed();

      using (var stream = new ChunkedMemoryStream()) {
        Save(stream);

        stream.Position = 0L;

        var reader = new StreamReader(stream, Charsets.ISO8859_1);

        return reader.ReadToEnd();
      }
    }

    private bool disposed = false;
    private /*readonly*/ MimeMessage parentPart;
    private /*readonly*/ MimeHeaderCollection headers;
    private Stream content;
    private /*readonly*/ MimeType mimeType = DefaultMimeType;
    private /*readonly*/ Encoding charset = DefaultCharset;
    private /*readonly*/ string boundary = null;
    private /*readonly*/ ContentTransferEncodingMethod transferEncoding = ContentTransferEncodingMethod.SevenBit;
    private /*readonly*/ MimeMessageDisposition disposition = MimeMessageDisposition.None;
    private /*readonly*/ ReadOnlyCollection<MimeMessage> subParts;
    private /*readonly*/ MimeMessage encapsulatedMessage = null;
  }
}