using System;
using System.IO;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class AttachedFileTests {
    [Test]
    public void TestCreateFrom()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: multipart/mixed; boundary=""******""

multipart message body
--******
Content-Type: text/plain
Content-Disposition: attachment; filename=""foo.txt""

part1
--******
Content-Type: text/plain
Content-Disposition: attachment

part2
--******--
"
      );
      const string directory = "attachedfile";

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        var attachedFile1 = AttachedFile.CreateFrom(mime.SubParts[0]);

        Assert.AreEqual("foo.txt", attachedFile1.FileName);
        Assert.AreEqual("part1", attachedFile1.ReadContentAsText());

        var file1 = attachedFile1.Save(directory);

        Assert.IsTrue(file1.Exists, "file #1 exists");
        Assert.AreEqual("foo.txt", file1.Name);
        Assert.AreEqual("part1", File.ReadAllText(file1.FullName), "save file content #1");

        file1.Delete();

        var attachedFile2 = AttachedFile.CreateFrom(mime.SubParts[1]);

        Assert.IsNull(attachedFile2.FileName);
        Assert.AreEqual("part2", attachedFile2.ReadContentAsText());

        var file2 = attachedFile2.Save(directory);

        Assert.IsTrue(file2.Exists, "file #2 exists");
        Assert.AreEqual("2.txt", file2.Name);
        Assert.AreEqual("part2", File.ReadAllText(file2.FullName), "save file content #1");

        file2.Delete();
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestReadContent()
    {
      var mail = Mail.LoadMail(@"Content-Type: multipart/mixed;
  boundary=""==boundary==""

multipart message body

--==boundary==
Content-Disposition: attachment
Content-Type: text/html

<html><body><p>attached part</p></body></html>
--==boundary==--");

      var attached = mail.AttachedFiles.First();

      Assert.AreEqual("<html><body><p>attached part</p></body></html>",
                      attached.ReadContentAsText());
    }

    private AttachedFile GetAttachedFile(string contentDisposition)
    {
      var mail = Mail.LoadMail(@"Content-Type: multipart/mixed;
  boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
" + contentDisposition + @"

<html><body><p>attached part</p></body></html>
--==boundary==--");

      return mail.AttachedFiles.First();
    }

    [Test]
    public void TestPropertiesParametersNotExist()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment;
Content-Type: text/html");

      Assert.IsNull(attachedFile.FileName);
      Assert.IsNull(attachedFile.CreationDate);
      Assert.IsNull(attachedFile.ModificationDate);
      Assert.IsNull(attachedFile.ReadDate);
      Assert.IsNull(attachedFile.Size);
    }

    [Test]
    public void TestPropertiesContentDispositionHeaderNotExist()
    {
      var attachedFile = GetAttachedFile(@"Content-Type: image/jpeg");

      Assert.IsNull(attachedFile.FileName);
      Assert.IsNull(attachedFile.CreationDate);
      Assert.IsNull(attachedFile.ModificationDate);
      Assert.IsNull(attachedFile.ReadDate);
      Assert.IsNull(attachedFile.Size);
    }

    [Test]
    public void TestFileNameContentDispositionFilenameParam()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=foo.html
Content-Type: text/html");

      Assert.AreEqual("foo.html", attachedFile.FileName);
    }

    [Test]
    public void TestFileNameContentTypeNameParam()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment
Content-Type: text/html; name=foo.html");

      Assert.AreEqual("foo.html", attachedFile.FileName);
    }

    [Test]
    public void TestFileNameBothContentDispositionContentTypeHaveParam()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=foo.html
Content-Type: text/html; name=bar.html");

      Assert.AreEqual("foo.html", attachedFile.FileName);
    }

    [Test]
    public void TestFileNameBothContentDispositionContentTypeNotHaveParam()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment
Content-Type: text/html");

      Assert.IsNull(attachedFile.FileName);
    }

    [Test]
    public void TestModificationDate()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    modification-date=""Wed, 12 Feb 1997 16:29:51 -0500""
Content-Type: image/jpeg");

      Assert.AreEqual(new DateTimeOffset(1997, 2, 12, 16, 29, 51, new TimeSpan(-5, 0, 0)),
                      attachedFile.ModificationDate);
    }

    [Test]
    public void TestModificationDateInvalidFormat()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    modification-date=""invalid datetime value""
Content-Type: image/jpeg");

      Assert.IsNull(attachedFile.ModificationDate);
    }

    [Test]
    public void TestCreationDate()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    creation-date=""Wed, 12 Feb 1997 16:29:51 -0500""
Content-Type: image/jpeg");

      Assert.AreEqual(new DateTimeOffset(1997, 2, 12, 16, 29, 51, new TimeSpan(-5, 0, 0)),
                      attachedFile.CreationDate);
    }


    [Test]
    public void TestCreationDateInvalidFormat()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    creation-date=""invalid datetime value""
Content-Type: image/jpeg");

      Assert.IsNull(attachedFile.CreationDate);
    }

    [Test]
    public void TestReadDate()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    read-date=""Wed, 12 Feb 1997 16:29:51 -0500""
Content-Type: image/jpeg");

      Assert.AreEqual(new DateTimeOffset(1997, 2, 12, 16, 29, 51, new TimeSpan(-5, 0, 0)),
                      attachedFile.ReadDate);
    }

    [Test]
    public void TestReadDateInvalidFormat()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    read-date=""invalid datetime value""
Content-Type: image/jpeg");

      Assert.IsNull(attachedFile.ReadDate);
    }

    [Test]
    public void TestSize()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    size=128
Content-Type: image/jpeg");

      Assert.AreEqual(128, attachedFile.Size);
    }

    [Test]
    public void TestSizeInvalidValue()
    {
      var attachedFile = GetAttachedFile(@"Content-Disposition: attachment; filename=genome.jpeg;
    size=foo
Content-Type: image/jpeg");

      Assert.IsNull(attachedFile.Size);
    }

    [Test]
    public void TestSave()
    {
      const string filename = "foo.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment; filename=""" + filename + @"""
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (File.Exists(filename))
          File.Delete(filename);

        Assert.AreEqual(filename, attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save();

        Assert.IsTrue(File.Exists(filename), "saved file exists");
        Assert.AreEqual(filename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (File.Exists(filename))
          File.Delete(filename);
      }
    }

    [Test]
    public void TestSaveTargetDirectorySpecified()
    {
      const string directory = "attachedfile";
      const string filename = "foo.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment; filename=""" + filename + @"""
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Assert.AreEqual(filename, attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save(directory);

        Assert.IsTrue(Directory.Exists(directory), "directory created");
        Assert.IsTrue(File.Exists(Path.Combine(directory, filename)), "saved file exists");
        Assert.AreEqual(filename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveTargetDirectorySpecifiedArgumentNull()
    {
      const string filename = "foo.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment; filename=""" + filename + @"""
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      Assert.Throws<ArgumentNullException>(delegate {
        attachedFile.Save(null);
      });
    }

    [Test]
    public void TestSaveTargetDirectorySpecifiedArgumentEmpty()
    {
      const string filename = "foo.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment; filename=""" + filename + @"""
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      Assert.Throws<ArgumentException>(delegate {
        attachedFile.Save(string.Empty);
      });
    }

    [Test]
    public void TestSaveFileNameStartsWithDirectory()
    {
      const string filename = "foo.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment; filename=""temp/directory/" + filename + @"""
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (File.Exists(filename))
          File.Delete(filename);

        Assert.AreEqual("temp/directory/" + filename, attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save();

        Assert.IsTrue(File.Exists(filename), "saved file exists");
        Assert.AreEqual(filename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (File.Exists(filename))
          File.Delete(filename);
      }
    }

    [Test]
    public void TestSaveGetDefaultFileNameFallback2_FromContentTypeFileParam()
    {
      const string directory = "attachedfile";
      const string filename = "foo.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain; name=""temp/directory/" + filename + @"""

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Assert.AreEqual("temp/directory/" + filename, attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save(directory);

        Assert.IsTrue(Directory.Exists(directory), "directory created");
        Assert.IsTrue(File.Exists(Path.Combine(directory, filename)), "saved file exists");
        Assert.AreEqual(filename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveGetDefaultFileNameFallback3_Subject()
    {
      const string directory = "attachedfile";
      const string expectedFilename = "添付ファイル.1.txt";
      const string expectedContent = "content";

      // Subject = "/添付ファイル/"
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""
Subject: =?iso-2022-jp?q?/=1B$BE:IU%U%!%$%k=1B(B/?=

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Assert.IsNull(attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save(directory);

        Assert.IsTrue(Directory.Exists(directory), "directory created");
        Assert.IsTrue(File.Exists(Path.Combine(directory, expectedFilename)), "saved file exists");
        Assert.AreEqual(expectedFilename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveGetDefaultFileNameFallback4_MessageID()
    {
      const string directory = "attachedfile";
      var expectedFilename = Runtime.IsRunningOnWindows ? "20121006214441.foobar@example.jp.1.txt" : "<20121006214441.foobar@example.jp>.1.txt";
      const string expectedContent = "content";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""
Message-ID: <20121006214441.foobar@example.jp>

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Assert.IsNull(attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save(directory);

        Assert.IsTrue(Directory.Exists(directory), "directory created");
        Assert.IsTrue(File.Exists(Path.Combine(directory, expectedFilename)), "saved file exists");
        Assert.AreEqual(expectedFilename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveGetDefaultFileNameFallback5_PartSpecifierOnly()
    {
      const string directory = "attachedfile";
      const string expectedFilename = "1.txt";
      const string expectedContent = "content";
      
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

" + expectedContent + @"
--==boundary==--";
      var attachedFile = Mail.LoadMail(input).AttachedFiles.First();

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Assert.IsNull(attachedFile.FileName, "FileName property value");

        var file = attachedFile.Save(directory);

        Assert.IsTrue(Directory.Exists(directory), "directory created");
        Assert.IsTrue(File.Exists(Path.Combine(directory, expectedFilename)), "saved file exists");
        Assert.AreEqual(expectedFilename, file.Name);

        Assert.AreEqual(expectedContent, File.ReadAllText(file.FullName));
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveMultipleAttachmentsWithPartSpecifier()
    {
      const string directory = "attachedfile";

      // Subject = "/添付ファイル/"
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""
Subject: =?iso-2022-jp?q?/=1B$BE:IU%U%!%$%k=1B(B/?=

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: message/rfc822

Content-Type: text/plain

part1
--==boundary==
Content-Disposition: attachment; filename=part2
Content-Type: text/plain

part2
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

nested part
--==nested==
Content-Disposition: attachment
Content-Type: image/bmp

part3.1
--==nested==
Content-Disposition: attachment
Content-Type: application/rtf

part3.2
--==nested==--
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var attachedFiles = Mail.LoadMail(input).AttachedFiles;

      Assert.AreEqual(4, attachedFiles.Count, "attached file count");

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        for (var index = 0; index < 4; index++) {
          attachedFiles[index].Save(directory);
        }

        Assert.IsTrue(Directory.Exists(directory), "directory created");
        Assert.AreEqual(4, Directory.GetFiles(directory).Length, "saved file counts");

        var filename = Path.Combine(directory, "添付ファイル.1.eml");

        Assert.IsTrue(File.Exists(filename), "saved file exists #0");
        Assert.AreEqual("Content-Type: text/plain\r\n\r\npart1", File.ReadAllText(filename), "content #0");

        filename = Path.Combine(directory, "part2");

        Assert.IsTrue(File.Exists(filename), "saved file exists #1");
        Assert.AreEqual("part2", File.ReadAllText(filename), "content #1");

        filename = Path.Combine(directory, "添付ファイル.3.1.bmp");

        Assert.IsTrue(File.Exists(filename), "saved file exists #2");
        Assert.AreEqual("part3.1", File.ReadAllText(filename), "content #2");

        filename = Path.Combine(directory, "添付ファイル.3.2.rtf");

        Assert.IsTrue(File.Exists(filename), "saved file exists #3");
        Assert.AreEqual("part3.2", File.ReadAllText(filename), "content #3");
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }
  }
}

