using System;
using System.IO;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MailTests {
    [Test]
    public void TestProperties()
    {
      var input = @"MIME-Version: 1.0
Return-Path: <return-path@gmail.com>
Received: from foo by bar with buz
Message-Id: <650A339D.4500208@example.com>
Date: Sun, 16 Sep 2012 16:45:11 +0900
From: from-address <from@example.com>
To: to-address <to@example.com>
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=
In-Reply-To: <CpN_A3fDtNWvrvegSSVUHpQcw@example.com>
References: <CpN_A3fDtNWvrvegSSVUHpQcw@example.com>
User-Agent: Mozilla/5.0 Thunderbird/14.0
Reply-To: reply-to-address <reply-to@example.com>
Sender: sender-address <sender@example.com>
CC: cc1@example.com, cc2@example.com
BCC: bcc1@example.com, bcc2@example.com
Content-Type: multipart/alternative;
  boundary=""==boundary==""
Content-Transfer-Encoding: 7bit

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--";

      var mail = Mail.LoadMail(input);

      Assert.IsNotNull(mail.BaseMessage);
      Assert.AreEqual(17, mail.Headers.Count);
      Assert.AreEqual(MimeType.MultipartAlternative, mail.MimeType);
      Assert.AreEqual(MimeMessageDisposition.None, mail.Disposition);

      Assert.AreEqual("件名", mail.Subject, "Subject");
      Assert.AreEqual("from-address <from@example.com>", mail.From, "From");
      Assert.AreEqual("to-address <to@example.com>", mail.To, "To");
      Assert.AreEqual("Sun, 16 Sep 2012 16:45:11 +0900", mail.DateString, "DateString");
      Assert.AreEqual(new DateTimeOffset(2012, 9, 16, 16, 45, 11, new TimeSpan(+9, 0, 0)), mail.Date, "Date");
      Assert.AreEqual("<650A339D.4500208@example.com>", mail.MessageId, "MessageId");
      Assert.AreEqual("<CpN_A3fDtNWvrvegSSVUHpQcw@example.com>", mail.InReplyTo, "InReplyTo");
      Assert.AreEqual("<CpN_A3fDtNWvrvegSSVUHpQcw@example.com>", mail.References, "References");
      Assert.AreEqual("Mozilla/5.0 Thunderbird/14.0", mail.UserAgent, "UserAgent");
      Assert.AreEqual("reply-to-address <reply-to@example.com>", mail.ReplyTo, "ReplyTo");
      Assert.AreEqual("sender-address <sender@example.com>", mail.Sender, "Sender");
      //Assert.AreEqual("cc1@example.com, cc2@example.com", mail.CC, "CC");
      //Assert.AreEqual("bcc1@example.com, bcc2@example.com", mail.Bcc, "Bcc");

      Assert.AreEqual(2, mail.SubParts.Count, "SubParts.Count");

      CollectionAssert.IsEmpty(mail.AttachmentParts, "AttachmentParts");
      CollectionAssert.IsEmpty(mail.InlineMessageParts, "InlineMessageParts");
    }

    [Test]
    [TestCase("Fri, 27 Jun 2014 09:31:48 +0900")]
    [TestCase("Fri, 27 Jun 2014 09:31:48 +0900 (JST)")]
    [TestCase("Fri, 27 Jun 2014 09:31:48 +0900 (GMT)")]
    public void TestDate(string dateString)
    {
      var input = @"MIME-Version: 1.0
Date: " + dateString + @"
Content-Type: text/plain

body
";

      var mail = Mail.LoadMail(input);

      Assert.AreEqual(new DateTimeOffset(2014, 6, 27, 09, 31, 48, new TimeSpan(+9, 0, 0)), mail.Date, "Date");
    }

    [Test]
    public void TestDate_WithExtraString()
    {
      var input = @"MIME-Version: 1.0
Date: Fri, 27 Jun 2014 09:31:48 +0900 extra string
Content-Type: text/plain

body
";
      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri, 27 Jun 2014 09:31:48 +0900 extra string", mail.DateString, "DateString");

      if (Runtime.IsRunningOnMono)
        Assert.AreEqual(new DateTimeOffset(2014, 6, 27, 09, 31, 48, new TimeSpan(+9, 0, 0)), mail.Date, "Date");
      else
        Assert.IsNull(mail.Date, "Date");
    }

    [Test]
    public void TestDate_WithComment()
    {
      var input = @"MIME-Version: 1.0
Date: Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002
  12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002\r\n  12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)",
                      mail.DateString,
                      "DateString");
      Assert.AreEqual(new DateTimeOffset(2002, 3, 15, 12, 32, 23, new TimeSpan(+9, 0, 0)), mail.Date, "Date");
    }

    [Test]
    public void TestDispose()
    {
      var input = @"MIME-Version: 1.0";

      var mail = Mail.LoadMail(input);

      mail.Dispose();

      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.BaseMessage));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Headers));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.SubParts));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.MimeType));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Disposition));

      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Subject), "Subject");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.From), "From");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.To), "To");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Date), "Date");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.DateString), "Date");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.MessageId), "MessageId");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.InReplyTo), "InReplyTo");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.References), "References");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.UserAgent), "UserAgent");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.ReplyTo), "ReplyTo");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Sender), "Sender");
      //Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.CC), "CC");
      //Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Bcc), "Bcc");

      Assert.Throws<ObjectDisposedException>(() => mail.AttachmentParts.ToArray(), "AttachmentParts");
      Assert.Throws<ObjectDisposedException>(() => mail.InlineMessageParts.ToArray(), "InlineMessageParts");

      Assert.Throws<ObjectDisposedException>(() => mail.GetHeaderValueOf("Content-Type"));

      Assert.DoesNotThrow(() => mail.Dispose(), "re-Dispose");
    }

    [Test]
    public void TestGetHeaderValueOf()
    {
      var input = @"MIME-Version: 1.0
Subject: foo
Subject: bar";

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("foo", mail.GetHeaderValueOf("SUBJECT"));
      Assert.IsNull(mail.GetHeaderValueOf("X-NonExistent"));
    }

    [Test]
    public void TestGetHeaderValueOf_WithParameter()
    {
      var input = @"MIME-Version: 1.0
Content-Type: text/plain; name=foo.txt";

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("text/plain; name=foo.txt", mail.GetHeaderValueOf("Content-Type"));
      Assert.AreEqual("text/plain; name=foo.txt", mail.GetHeaderValueOf("Content-Type", true));
      Assert.AreEqual("text/plain", mail.GetHeaderValueOf("Content-Type", false));
    }

    [Test]
    public void TestGetHeaderValueOf_RemoveComment()
    {
      var input = @"MIME-Version: 1.0
Date: Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002
  12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002\r\n  12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)",
                      mail.GetHeaderValueOf("Date"));
      Assert.AreEqual("Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002 12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)",
                      mail.GetHeaderValueOf("Date", false, false));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900",
                      mail.GetHeaderValueOf("Date", false, true));
    }

    [Test]
    public void TestGetHeaderValueOf_RemoveComment_QuotedPair_1()
    {
      var input = @"MIME-Version: 1.0
Date: Fri, 15 Mar 2002 12:32:23 +0900 \(JST\) extratext
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 \\(JST\\) extratext",
                      mail.GetHeaderValueOf("Date"));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 \\(JST\\) extratext",
                      mail.GetHeaderValueOf("Date", false, false));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 \\(JST\\) extratext",
                      mail.GetHeaderValueOf("Date", false, true));
    }

    [Test]
    public void TestGetHeaderValueOf_RemoveComment_QuotedPair_2()
    {
      var input = @"MIME-Version: 1.0
Date: Fri, 15 Mar 2002 12:32:23 +0900 (JST\)) extratext
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 (JST\\)) extratext",
                      mail.GetHeaderValueOf("Date"));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 (JST\\)) extratext",
                      mail.GetHeaderValueOf("Date", false, false));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 extratext",
                      mail.GetHeaderValueOf("Date", false, true));
    }

    [Test]
    public void TestGetHeaderValueOf_RemoveComment_QuotedPair_3()
    {
      var input = @"MIME-Version: 1.0
Date: Fri, 15 Mar 2002 12:32:23 +0900 \
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 \\",
                      mail.GetHeaderValueOf("Date"));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 \\",
                      mail.GetHeaderValueOf("Date", false, false));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 \\",
                      mail.GetHeaderValueOf("Date", false, true));
    }

    [Test]
    public void TestGetHeaderValueOf_RemoveComment_NestedComment()
    {
      var input = @"MIME-Version: 1.0
Date: Fri, 15 Mar 2002 12:32:23 +0900 (JST(Japan Standard time)) extratext
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 (JST(Japan Standard time)) extratext",
                      mail.GetHeaderValueOf("Date"));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 (JST(Japan Standard time)) extratext",
                      mail.GetHeaderValueOf("Date", false, false));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 extratext",
                      mail.GetHeaderValueOf("Date", false, true));
    }

    [Test]
    public void TestGetHeaderValueOf_RemoveComment_UnmatchNest()
    {
      var input = @"MIME-Version: 1.0
Date: Fri, 15 Mar 2002 12:32:23 +0900 (JST)) extratext
Content-Type: text/plain

body
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 (JST)) extratext",
                      mail.GetHeaderValueOf("Date"));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 (JST)) extratext",
                      mail.GetHeaderValueOf("Date", false, false));
      Assert.AreEqual("Fri, 15 Mar 2002 12:32:23 +0900 extratext",
                      mail.GetHeaderValueOf("Date", false, true));
    }

    [Test]
    public void TestAttachedFiles()
    {
      var input = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/css

/*part#0*/
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/css

/*nested#0*/
--==nested==
Content-Type:text/css

/*nested#1*/
--==nested==--
--==boundary==
Content-Type: text/css

/*part#1*/
--==boundary==--";

      var mail = Mail.LoadMail(input);

      Assert.IsNotEmpty(mail.AttachedFiles);
      Assert.AreEqual(4, mail.AttachedFiles.Count);
      Assert.AreSame(mail.AttachedFiles, mail.AttachedFiles);

      var attachedFiles = mail.AttachedFiles;

      Assert.AreSame(mail.BaseMessage.SubParts[0], attachedFiles[0].BaseMessage, "AttachedFiles[0].BaseMessage");
      Assert.AreSame(mail.BaseMessage.SubParts[1].SubParts[0], attachedFiles[1].BaseMessage, "AttachedFiles[1].BaseMessage");
      Assert.AreSame(mail.BaseMessage.SubParts[1].SubParts[1], attachedFiles[2].BaseMessage, "AttachedFiles[2].BaseMessage");
      Assert.AreSame(mail.BaseMessage.SubParts[2], attachedFiles[3].BaseMessage, "AttachedFiles[3].BaseMessage");

      Assert.AreSame(mail, attachedFiles[0].Mail, "AttachedFiles[0].Mail");
      Assert.AreSame(mail, attachedFiles[1].Mail, "AttachedFiles[1].Mail");
      Assert.AreSame(mail, attachedFiles[2].Mail, "AttachedFiles[2].Mail");
      Assert.AreSame(mail, attachedFiles[3].Mail, "AttachedFiles[3].Mail");

      var attachedFiles2 = mail.AttachedFiles;

      for (var i = 0; i < attachedFiles.Count; i++) {
        Assert.AreSame(attachedFiles[i], attachedFiles2[i], "get again AttachedFiles[{0}]", i);
      }
    }

    [Test]
    public void TestIsAttachedFile()
    {
      var input = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Disposition: attachment
Content-Type: text/plain
Content-Description: attachment part

part1
--==boundary==
Content-Disposition: inline
Content-Type: text/plain

part2
--==boundary==
Content-Type: text/plain

part3
--==boundary==
Content-Type: multipart/alternative; boundary=""******""

part4
--******
Content-Type: text/plain

part4.1
--******
Content-Type: text/html

part4.2
--******
Content-Disposition: attachment
Content-Type: text/html
Content-Description: attachment part

part4.3
--******
Content-Type: image/jpeg

part4.4
--******
Content-Type: application/octet-stream

part4.5
--******--
--==boundary==
Content-Type: image/jpg
Content-Description: attachment part

part5
--==boundary==
Content-Disposition: inline
Content-Type: image/jpg

part6
--==boundary==
Content-Type: message/partial

part7
--==boundary==
Content-Disposition: attachment
Content-Type: message/partial
Content-Description: attachment part

part8
--==boundary==--";

      var mail = Mail.LoadMail(input);

      Assert.AreEqual(4, mail.AttachedFiles.Count);

      Assert.AreEqual("part1", mail.AttachedFiles[0].ReadContentAsText(), "AttachedFiles[0]");
      Assert.AreEqual("part4.3", mail.AttachedFiles[1].ReadContentAsText(), "AttachedFiles[1]");
      Assert.AreEqual("part5", mail.AttachedFiles[2].ReadContentAsText(), "AttachedFiles[2]");
      Assert.AreEqual("part8", mail.AttachedFiles[3].ReadContentAsText(), "AttachedFiles[3]");
    }

    [Test]
    public void TestSaveAttachedFilesToDirectory()
    {
      const string directory = "attachedfile";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""
Subject: =?iso-2022-jp?q?/=1B$BE:IU%U%!%$%k=1B(B/?=

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: message/rfc822

Content-Type: text/plain

part1
--==boundary==
Content-Disposition: attachment; filename=part2
Content-Type: text/plain

part2
--==boundary==
Content-Disposition: inline; filename=part3
Content-Type: text/plain

part3
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

nested part
--==nested==
Content-Disposition: attachment
Content-Type: image/bmp; name=test.bmp

part4.1
--==nested==
Content-Disposition: attachment
Content-Type: application/rtf

part4.2
--==nested==--
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var mail = Mail.LoadMail(input);

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        var files = mail.SaveAttachedFilesToDirectory(directory);

        Assert.AreEqual(4, files.Length);

        Assert.AreEqual("添付ファイル.1.eml", files[0].Name, "file name #0");
        StringAssert.EndsWith(Path.Combine(directory, "添付ファイル.1.eml"), files[0].FullName, "file full name #0");
        Assert.AreEqual("Content-Type: text/plain\r\n\r\npart1", File.ReadAllText(files[0].FullName), "content #0");

        Assert.AreEqual("part2", files[1].Name, "file name #1");
        StringAssert.EndsWith(Path.Combine(directory, "part2"), files[1].FullName, "file full name #1");
        Assert.AreEqual("part2", File.ReadAllText(files[1].FullName), "content #1");

        Assert.AreEqual("test.bmp", files[2].Name, "file name #2");
        StringAssert.EndsWith(Path.Combine(directory, "test.bmp"), files[2].FullName, "file full name #2");
        Assert.AreEqual("part4.1", File.ReadAllText(files[2].FullName), "content #2");

        Assert.AreEqual("添付ファイル.4.2.rtf", files[3].Name, "file name #3");
        StringAssert.EndsWith(Path.Combine(directory, "添付ファイル.4.2.rtf"), files[3].FullName, "file full name #3");
        Assert.AreEqual("part4.2", File.ReadAllText(files[3].FullName), "content #3");
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveAttachedFilesToDirectoryCurrentDirectory()
    {
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

part
--==boundary==--";
      var mail = Mail.LoadMail(input);

      FileInfo[] files = null;

      try {
        files = mail.SaveAttachedFilesToDirectory(".");

        Assert.AreEqual(1, files.Length, "file count");
        Assert.AreEqual("1.txt", files[0].Name, "file name");
      }
      finally {
        if (files != null) {
          foreach (var file in files) {
            file.Delete();
          }
        }
      }
    }

    [Test]
    public void TestSaveAttachedFilesToDirectoryArgumentNull()
    {
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

part
--==boundary==--";
      var mail = Mail.LoadMail(input);

      Assert.Throws<ArgumentNullException>(delegate {
        mail.SaveAttachedFilesToDirectory(null);
      });
    }

    [Test]
    public void TestSaveAttachedFilesToDirectoryArgumentEmpty()
    {
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

part
--==boundary==--";
      var mail = Mail.LoadMail(input);
      
      Assert.Throws<ArgumentException>(delegate {
        mail.SaveAttachedFilesToDirectory(string.Empty);
      });
    }

    [Test]
    public void TestSaveAttachedFilesToDirectoryDisposed()
    {
      const string directory = "attachedfile";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

part
--==boundary==--";
      var mail = Mail.LoadMail(input);

      if (Directory.Exists(directory))
        Directory.Delete(directory, true);

      mail.Dispose();

      Assert.Throws<ObjectDisposedException>(delegate {
        mail.SaveAttachedFilesToDirectory(directory);
      });

      Assert.IsFalse(Directory.Exists(directory), "directory not created");
    }

    [Test]
    public void TestSaveAttachedFilesAs()
    {
      const string directory = "attachedfile";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""
Subject: =?iso-2022-jp?q?/=1B$BE:IU%U%!%$%k=1B(B/?=

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: message/rfc822

Content-Type: text/plain

part1
--==boundary==
Content-Disposition: attachment; filename=part2
Content-Type: text/plain

part2
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

nested part
--==nested==
Content-Disposition: attachment
Content-Type: image/bmp; name=test.bmp

part3.1
--==nested==
Content-Disposition: attachment
Content-Type: application/rtf

part3.2
--==nested==--
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var mail = Mail.LoadMail(input);

      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Directory.CreateDirectory(directory);

        var files = mail.SaveAttachedFilesAs(delegate(AttachedFile attachedFile) {
          return Path.Combine(directory, attachedFile.BaseMessage.PartSpecifier + ".eml");
        });

        Assert.AreEqual(4, files.Length);

        Assert.AreEqual("1.eml", files[0].Name, "file name #0");
        StringAssert.EndsWith(Path.Combine(directory, "1.eml"), files[0].FullName, "file full name #0");
        Assert.AreEqual("Content-Type: text/plain\r\n\r\npart1", File.ReadAllText(files[0].FullName), "content #0");
        
        Assert.AreEqual("2.eml", files[1].Name, "file name #1");
        StringAssert.EndsWith(Path.Combine(directory, "2.eml"), files[1].FullName, "file full name #1");
        Assert.AreEqual("part2", File.ReadAllText(files[1].FullName), "content #1");
        
        Assert.AreEqual("3.1.eml", files[2].Name, "file name #2");
        StringAssert.EndsWith(Path.Combine(directory, "3.1.eml"), files[2].FullName, "file full name #2");
        Assert.AreEqual("part3.1", File.ReadAllText(files[2].FullName), "content #2");
        
        Assert.AreEqual("3.2.eml", files[3].Name, "file name #3");
        StringAssert.EndsWith(Path.Combine(directory, "3.2.eml"), files[3].FullName, "file full name #3");
        Assert.AreEqual("part3.2", File.ReadAllText(files[3].FullName), "content #3");
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveAttachedFilesAsDirectoryNotExist()
    {
      const string directory = "attachedfile";

      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

part
--==boundary==--";
      var mail = Mail.LoadMail(input);
      try {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);

        Assert.Throws<DirectoryNotFoundException>(delegate {
          mail.SaveAttachedFilesAs(delegate(AttachedFile attachedFile) {
            return Path.Combine(directory, attachedFile.BaseMessage.PartSpecifier + ".eml");
          });
        });

        Assert.IsFalse(Directory.Exists(directory), "directory not created");
      }
      finally {
        if (Directory.Exists(directory))
          Directory.Delete(directory, true);
      }
    }

    [Test]
    public void TestSaveAttachedFilesAsDirectoryArgumentNull()
    {
      const string directory = "attachedfile";
      
      var input = @"Content-Type: multipart/mixed; boundary=""==boundary==""

This is a multi-part message in MIME format.

--==boundary==
Content-Disposition: attachment
Content-Type: text/plain

part
--==boundary==--";
      var mail = Mail.LoadMail(input);

      Assert.Throws<ArgumentNullException>(delegate {
        mail.SaveAttachedFilesAs(null);
      });
    }
  }
}