using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MimeHeaderTests {
    [Test]
    public void TestValueEmpty()
    {
      var input = @"Content-Type:
";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.IsEmpty(contentType.Value);
      CollectionAssert.IsEmpty(contentType.GetAllParameters());
    }

    [Test]
    public void TestClone()
    {
      var input = @"Content-Type: text/plain;
  foo=hoge;
  bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      var clone = contentType.Clone();

      Assert.AreNotSame(contentType, clone);
      Assert.AreEqual(contentType.Name, clone.Name, "Name");
      Assert.AreEqual(contentType.Value, clone.Value, "Name");
      Assert.AreSame(contentType.Charset, clone.Charset, "Charset");
      Assert.AreEqual(contentType.Encoding, clone.Encoding, "Encoding");

      Assert.AreEqual(contentType.GetParameter("foo"),
                      clone.GetParameter("foo"), "GetParameter #1");
      Assert.AreEqual(contentType.GetParameter("FOO"),
                      clone.GetParameter("FOO"), "GetParameter #2");
      Assert.AreEqual(contentType.GetParameter("bar"),
                      clone.GetParameter("bar"), "GetParameter #3");
      Assert.AreEqual(contentType.GetParameter("baz"),
                      clone.GetParameter("baz"), "GetParameter #4");
    }

    [TestCase("Content-Type:text/plain\r\n")]
    [TestCase("Content-Type\t:\ttext/plain  \n")]
    [TestCase("CONTENT-TYPE :  text/plain\n foo=hoge; bar=huga;  \n")]
    [TestCase("Content-Type:\n")]
    [TestCase("Content-Type:")]
    public void TestToString(string expected)
    {
      Assert.AreEqual(expected,
                      MimeMessage.LoadMessage(expected).Headers["Content-Type"].ToString());
    }

    [Test]
    public void TestGetParameter()
    {
      var input = @"Content-Type: text/plain;
  foo=hoge;
  bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("hoge", contentType.GetParameter("foo"));
      Assert.AreEqual("hoge", contentType.GetParameter("FOO"));
      Assert.AreEqual("huga", contentType.GetParameter("bar"));
      Assert.AreEqual("huga", contentType.GetParameter("BAR"));
      Assert.IsNull(contentType.GetParameter("baz"));
    }

    [Test]
    public void TestGetAllParameters()
    {
      var input = @"Content-Type: text/plain;
  foo=hoge;
  bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      var parameters = contentType.GetAllParameters();

      Assert.AreEqual(2, parameters.Length);

      Assert.AreEqual("hoge", parameters.Where(p => p.Key == "foo").First().Value);
      Assert.AreEqual("huga", parameters.Where(p => p.Key == "bar").First().Value);
    }

    [Test]
    public void TestForEachParameters()
    {
      var input = @"Content-Type: text/plain;
  foo=hoge;
  bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      var dic = new Dictionary<string, string>();

      contentType.ForEachParameter(delegate(KeyValuePair<string, string> pair) {
        dic[pair.Key] = pair.Value;
      });

      Assert.AreEqual(2, dic.Count);

      Assert.AreEqual("hoge", dic["foo"]);
      Assert.AreEqual("huga", dic["bar"]);
    }

    [Test]
    public void TestHasParameter()
    {
      var input = @"Content-Type: text/plain;
  foo=hoge;
  bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.IsTrue(contentType.HasParameter("foo"));
      Assert.IsTrue(contentType.HasParameter("FOO"));
      Assert.IsTrue(contentType.HasParameter("bar"));
      Assert.IsTrue(contentType.HasParameter("BAR"));
      Assert.IsFalse(contentType.HasParameter("baz"));
    }

    [Test]
    public void TestGetValueWithoutParameter()
    {
      var input = @"Content-Type: text/plain;
  foo=hoge;
  bar=huga
Content-Disposition: inline";
      var mime = MimeMessage.LoadMessage(input);
      var contentType = mime.Headers["Content-Type"];
      var contentDisposition = mime.Headers["Content-Disposition"];

      Assert.AreEqual("text/plain", contentType.GetValueWithoutParameter());
      Assert.AreEqual("inline", contentDisposition.GetValueWithoutParameter());
    }

    [Test]
    public void TestGetValueWithoutParameterFieldHasEmptyParameter()
    {
      var input = @"Content-Type: text/plain
  ;
Content-Disposition: inline;  
";
      var mime = MimeMessage.LoadMessage(input);
      var contentType = mime.Headers["Content-Type"];
      var contentDisposition = mime.Headers["Content-Disposition"];

      Assert.AreEqual("text/plain", contentType.GetValueWithoutParameter());
      Assert.AreEqual("inline", contentDisposition.GetValueWithoutParameter());
    }

    [Test]
    public void TestParseParameter()
    {
      var input = "Content-Type:text/plain;foo=hoge;bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("text/plain", contentType.GetValueWithoutParameter());
      Assert.AreEqual("hoge", contentType.GetParameter("foo"));
      Assert.AreEqual("huga", contentType.GetParameter("bar"));
    }

    [Test]
    public void TestParseParameterContainsLeadingAndTrailingWhitespace()
    {
      var input = "Content-Type: text/plain\t;\tfoo=hoge ; bar=huga";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("text/plain", contentType.GetValueWithoutParameter());
      Assert.AreEqual("hoge", contentType.GetParameter("foo"));
      Assert.AreEqual("huga", contentType.GetParameter("bar"));
    }

    [Test]
    public void TestParseParameterEndsWithDelimiter1()
    {
      var input = "Content-Type:text/plain; foo=hoge;";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("text/plain", contentType.GetValueWithoutParameter());
      Assert.AreEqual("hoge", contentType.GetParameter("foo"));
    }

    [Test]
    public void TestParseParameterEndsWithDelimiter2()
    {
      var input = "Content-Type:text/plain; foo=hoge; bar=huga;  ";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("text/plain", contentType.GetValueWithoutParameter());
      Assert.AreEqual("hoge", contentType.GetParameter("foo"));
      Assert.AreEqual("huga", contentType.GetParameter("bar"));
    }

    [Test]
    public void TestParseParameterWithoutParameters()
    {
      var input = @"Received: from example.com
  (example.com [111.111.11.111])
  by mx.example.net with SMTP id xxxyyyzzz;
  Tue, 29 Mar 2005 15:11:47 -0800 (PST)".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var expectedValue = @"from example.com
  (example.com [111.111.11.111])
  by mx.example.net with SMTP id xxxyyyzzz;
  Tue, 29 Mar 2005 15:11:47 -0800 (PST)".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var received = MimeMessage.LoadMessage(input).Headers["Received"];

      Assert.AreEqual(expectedValue, received.Value);
      Assert.AreEqual("from example.com (example.com [111.111.11.111]) by mx.example.net with SMTP id xxxyyyzzz; Tue, 29 Mar 2005 15:11:47 -0800 (PST)", received.GetValueWithoutParameter());
    }

    [Test]
    public void TestParseParameterQuoted()
    {
      var input = @"Content-Disposition: attachment;
  filename=""=?ISO-2022-JP?B?GyRCJF4kJCRhGyhCLnR4dA==?=""";
      var contentDisposition = MimeMessage.LoadMessage(input).Headers["Content-Disposition"];

      Assert.AreEqual("attachment", contentDisposition.GetValueWithoutParameter());
      Assert.AreEqual("まいめ.txt", contentDisposition.GetParameter("filename"));
    }

    [Test]
    public void TestParseParameterSectionedAndQuoted()
    {
      var input = @"Content-Type: message/external-body; access-type=URL;
  URL*0=""ftp://"";
  URL*1=""cs.utk.edu/pub/moore/bulk-mailer/bulk-mailer.tar""";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("message/external-body", contentType.GetValueWithoutParameter());
      Assert.AreEqual("URL", contentType.GetParameter("access-type"));
      Assert.AreEqual("ftp://cs.utk.edu/pub/moore/bulk-mailer/bulk-mailer.tar", contentType.GetParameter("URL"));
    }

    [Test]
    public void TestParseParameterExtended()
    {
      var input = @"Content-Type: application/x-stuff;
  title*=us-ascii'en-us'This%20is%20%2A%2A%2Afun%2A%2A%2A";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("application/x-stuff", contentType.GetValueWithoutParameter());
      Assert.AreEqual("This is ***fun***", contentType.GetParameter("title"));
    }

    [Test]
    public void TestParseParameterSectionedAndExtended1()
    {
      var input = @"Content-Type: application/x-stuff;
  title*0*=us-ascii'en'This%20is%20even%20more%20;
  title*1*=%2A%2A%2Afun%2A%2A%2A%20;
  title*2=""isn't it!""";
      var contentType = MimeMessage.LoadMessage(input).Headers["Content-Type"];

      Assert.AreEqual("application/x-stuff", contentType.GetValueWithoutParameter());
      Assert.AreEqual("This is even more ***fun*** isn't it!", contentType.GetParameter("title"));
    }

    [Test]
    public void TestParseParameterSectionedAndExtended2()
    {
      var input = @"Content-Disposition: attachment;
  filename*0*=iso-2022-jp'ja'%1B%24B%24%5B%244%24%5B%242%1B%28B;
  filename*1=.jpeg";
      var contentDisposition = MimeMessage.LoadMessage(input).Headers["Content-Disposition"];

      Assert.AreEqual("attachment", contentDisposition.GetValueWithoutParameter());
      Assert.AreEqual("ほごほげ.jpeg", contentDisposition.GetParameter("filename"));
    }

    [Test]
    public void TestParseParameterSectionedAndExtended3()
    {
      var input = @"Content-Disposition: inline;
  filename*0*=UTF-8''%E3%81%82%E3%81%84%E3%81%86%E3%81%88%E3%81%8A%55%54%46;
  filename*1*=%38%E3%81%AA%E3%83%95%E3%82%A1%E3%82%A4%E3%83%AB%E5%90%8D%2E;
  filename*2*=%74%78%74";

      var contentDisposition = MimeMessage.LoadMessage(input).Headers["Content-Disposition"];

      Assert.AreEqual("inline", contentDisposition.GetValueWithoutParameter());
      Assert.AreEqual("あいうえおUTF8なファイル名.txt", contentDisposition.GetParameter("filename"));
    }

    [Test]
    public void TestParseParameterParametersOnly()
    {
      var input = @"X-Foo: p1=foo; p2=bar; p3=baz";
      var header= MimeMessage.LoadMessage(input).Headers["X-Foo"];

      Assert.AreEqual(string.Empty, header.GetValueWithoutParameter());
      Assert.AreEqual("foo", header.GetParameter("p1"));
      Assert.AreEqual("bar", header.GetParameter("p2"));
      Assert.AreEqual("baz", header.GetParameter("p3"));
    }

    [Test]
    public void TestParseParameterParametersOnlySectioned()
    {
      var input = @"X-Foo: p*0=foo;
  p*1=bar;
  p*2=baz";
      var header= MimeMessage.LoadMessage(input).Headers["X-Foo"];

      Assert.AreEqual(string.Empty, header.GetValueWithoutParameter());
      Assert.AreEqual("foobarbaz", header.GetParameter("p"));
    }

    [Test]
    public void TestParseParameterMultipleContinuationLines()
    {
      var input = @"DKIM-Signature: a=rsa-sha256; d=example.net; s=brisbane;
    c=simple; q=dns/txt; i=@eng.example.net;
    t=1117574938; x=1118006938;
    h=from:to:subject:date;
    z=From:foo@eng.example.net|To:joe@example.com|
      Subject:demo=20run|Date:July=205,=202005=203:44:08=20PM=20-0700;
    bh=MTIzNDU2Nzg5MDEyMzQ1Njc4OTAxMjM0NTY3ODkwMTI=;
    b=dzdVyOfAKCdLXdJOc9G2q8LoXSlEniSbav+yuU4zGeeruD00lszZ
               VoG4ZHRNiYzR";
      var dkimsig = MimeMessage.LoadMessage(input).Headers["DKIM-Signature"];

      Assert.AreEqual(string.Empty, dkimsig.GetValueWithoutParameter());
      Assert.AreEqual("rsa-sha256", dkimsig.GetParameter("a"));
      Assert.AreEqual("example.net", dkimsig.GetParameter("d"));
      Assert.AreEqual("brisbane", dkimsig.GetParameter("s"));
      Assert.AreEqual("simple", dkimsig.GetParameter("c"));
      Assert.AreEqual("dns/txt", dkimsig.GetParameter("q"));
      Assert.AreEqual("@eng.example.net", dkimsig.GetParameter("i"));
      Assert.AreEqual("1117574938", dkimsig.GetParameter("t"));
      Assert.AreEqual("1118006938", dkimsig.GetParameter("x"));
      Assert.AreEqual("from:to:subject:date", dkimsig.GetParameter("h"));
      Assert.AreEqual("From:foo@eng.example.net|To:joe@example.com|Subject:demo=20run|Date:July=205,=202005=203:44:08=20PM=20-0700",
                      dkimsig.GetParameter("z"));
      Assert.AreEqual("MTIzNDU2Nzg5MDEyMzQ1Njc4OTAxMjM0NTY3ODkwMTI=", dkimsig.GetParameter("bh"));
      Assert.AreEqual("dzdVyOfAKCdLXdJOc9G2q8LoXSlEniSbav+yuU4zGeeruD00lszZVoG4ZHRNiYzR",
                      dkimsig.GetParameter("b"));
    }
  }
}

