using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MimeHeaderCollectionTests {
    [Test]
    public void TestIndexer()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content"
      );

      Assert.AreEqual(3, mime.Headers.Count);

      Assert.IsNotNull(mime.Headers[0]);
      Assert.IsNotNull(mime.Headers[1]);
      Assert.IsNotNull(mime.Headers[2]);

      Assert.AreSame(mime.Headers["MIME-Version"], mime.Headers[0]);
      Assert.AreSame(mime.Headers["Content-Type"], mime.Headers[1]);
      Assert.AreSame(mime.Headers["Subject"], mime.Headers[2]);

      Assert.Throws<ArgumentOutOfRangeException>(delegate {
        var header = mime.Headers[3];
      });

      Assert.IsNotNull(mime.Headers["Content-Type"]);
      Assert.AreSame(mime.Headers["Content-Type"], mime.Headers["CONTENT-TYPE"]);

      Assert.Throws<KeyNotFoundException>(delegate {
        var header = mime.Headers["non-existent"];
      });
    }

    [Test]
    public void TestIndexerHeadersContainSameName()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Received: #1
Received: #2
Subject: foo
Received: #3

content");

      Assert.IsNotNull(mime.Headers[0]);
      Assert.IsNotNull(mime.Headers[1]);
      Assert.IsNotNull(mime.Headers[2]);
      Assert.IsNotNull(mime.Headers[3]);
      Assert.IsNotNull(mime.Headers[4]);
      Assert.IsNotNull(mime.Headers["Received"]);

      Assert.AreEqual("#1", mime.Headers[1].Value);
      Assert.AreEqual("#2", mime.Headers[2].Value);
      Assert.AreEqual("#3", mime.Headers[4].Value);

      Assert.AreSame(mime.Headers["Received"], mime.Headers[1], "returns first header");

      Assert.AreEqual("MIME-Version", mime.Headers[0].Name);
      Assert.AreSame(mime.Headers["MIME-Version"], mime.Headers[0]);

      Assert.AreEqual("Subject", mime.Headers[3].Name);
      Assert.AreSame(mime.Headers["Subject"], mime.Headers[3]);
    }

    [Test]
    public void TestEnumerator()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Received: #1
Received: #2
Subject: foo
Received: #3

content");
      var expected = new KeyValuePair<string, string>[] {
        new KeyValuePair<string, string>("MIME-Version", "1.0"),
        new KeyValuePair<string, string>("Received", "#1"),
        new KeyValuePair<string, string>("Received", "#2"),
        new KeyValuePair<string, string>("Subject", "foo"),
        new KeyValuePair<string, string>("Received", "#3"),
      };
      var actual = mime.Headers.Select(h => new KeyValuePair<string, string>(h.Name, h.Value)).ToArray();

      Assert.AreEqual(expected.Length, actual.Length);

      for (var i = 0; i < expected.Length; i++) {
        Assert.AreEqual(expected[i].Key, actual[i].Key, "name #{0}", i);
        Assert.AreEqual(expected[i].Value, actual[i].Value, "value #{0}", i);
      }
    }

    [Test]
    public void TestContains()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");

      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.IsTrue(mime.Headers.Contains("CONTENT-TYPE"));
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.IsTrue(mime.Headers.Contains("subject"));
      Assert.IsFalse(mime.Headers.Contains("Date"));
    }

    [Test]
    public void TestTryGetHeader()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Received: #1
Received: #2
Subject: foo
Received: #3

content");

      MimeHeader header;

      Assert.IsTrue(mime.Headers.TryGetHeader("MIME-Version", out header), "#1");
      Assert.IsNotNull(header, "#1");

      Assert.IsTrue(mime.Headers.TryGetHeader("mime-version", out header), "#2");
      Assert.IsNotNull(header, "#2");

      Assert.IsTrue(mime.Headers.TryGetHeader("Received", out header), "#3");
      Assert.IsNotNull(header, "#3");
      Assert.AreEqual("#1", header.Value, "#3");

      Assert.IsFalse(mime.Headers.TryGetHeader("NonExistent", out header), "#4");
      Assert.IsNull(header, "#4");
    }

    [Test]
    public void TestGetHeaders()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Received: #1
Received: #2
Subject: foo
Received: #3

content");

      CollectionAssert.IsNotEmpty(mime.Headers.GetHeaders("MIME-Version"));
      Assert.AreEqual(1, mime.Headers.GetHeaders("MIME-Version").Count());
      Assert.AreEqual("1.0", mime.Headers.GetHeaders("MIME-Version").First().Value);
      Assert.AreEqual("1.0", mime.Headers.GetHeaders("mime-version").First().Value);

      CollectionAssert.IsEmpty(mime.Headers.GetHeaders("non-existent"));

      CollectionAssert.AllItemsAreUnique(mime.Headers.GetHeaders("Received"));
      Assert.AreEqual(3, mime.Headers.GetHeaders("Received").Count());
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#1").Count(), "#1");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#2").Count(), "#2");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#3").Count(), "#2");
    }
  }
}
