using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Formats.UUEncoding;

namespace Smdn.Formats.Mime {
  public abstract class MimeMessageTestsBase {
    public enum LoadMessageMode {
      OnMemory,
      FileStream,
    }

    protected void LoadMessage(string content, LoadMessageMode mode, Action<MimeMessage> action)
    {
      if (mode == LoadMessageMode.FileStream) {
        const string file = "message.eml";

        try {
          if (File.Exists(file))
            File.Delete(file);

          File.WriteAllText(file, content, Charsets.Latin1);

          using (var mime = MimeMessage.Load(file, false)) {
            if (mime.TransferEncoding != ContentTransferEncodingMethod.Unknown) {
              mime.ReadContent(delegate(Stream s) {
                if (mime.MimeType.TypeEqualsIgnoreCase("message") && !mime.MimeType.SubTypeEqualsIgnoreCase("partial")) {
                  Assert.IsInstanceOf<NonClosingStream>(s, "unwrapped stream (raw) #1");
                  var ncs = s as NonClosingStream;
                  Assert.IsInstanceOf<Stream>(ncs.InnerStream, "encapsulated message content stream #1");
                  Assert.AreSame(Stream.Null, ncs.InnerStream, "encapsulated message content stream #2");
                  return;
                }

                switch (mime.TransferEncoding) {
                  case ContentTransferEncodingMethod.Binary:
                  case ContentTransferEncodingMethod.EightBit:
                  case ContentTransferEncodingMethod.SevenBit: {
                    Assert.IsInstanceOf<NonClosingStream>(s, "unwrapped stream (raw) #1");
                    var ncs = s as NonClosingStream;
                    Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "unwrapped stream (raw) #2");
                    var ps = ncs.InnerStream as PartialStream;
                    Assert.IsInstanceOf<FileStream>(ps.InnerStream, "unwrapped stream (raw) #3");
                    var fs = ps.InnerStream as FileStream;
                    Assert.AreEqual(file, Path.GetFileName(fs.Name), "unwrapped stream file");
                    break;
                  }

                  case ContentTransferEncodingMethod.Base64:
                  case ContentTransferEncodingMethod.QuotedPrintable: {
                    Assert.IsInstanceOf<NonClosingStream>(s, "unwrapped stream (crypto) #1");
                    var ncs = s as NonClosingStream;
                    Assert.IsInstanceOf<System.Security.Cryptography.CryptoStream>(ncs.InnerStream, "unwrapped stream (crypto) #2");
                    break;
                  }

                  case ContentTransferEncodingMethod.UUEncode: {
                    Assert.IsInstanceOf<UUDecodingStream>(s, "unwrapped stream (uuencode) #1");
                    break;
                  }
                }
              });
            }

            action(mime);
          }
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      }
      else {
        using (var mime = MimeMessage.LoadMessage(content)) {
          action(mime);
        }
      }
    }
  }

  [TestFixture]
  public class MimeMessageTests : MimeMessageTestsBase {
    [Test]
    public void TestLoadMessageUnsupportedMimeVersion()
    {
      MimeMessage.LoadMessage(@"MIME-Version: 1.1
Content-Type: text/plain; charset=""iso-2022-jp""
Content-Transfer-Encoding: quoted-printable
");
    }

    [Test]
    public void TestLoadMessageMimeVersionNotExist()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain");

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
    }

    [Test]
    public void TestLoadMessageMimeVersionNotSupported()
    {
      MimeMessage.LoadMessage(@"MIME-Version: 1.1\r\n");
    }

    [Test]
    // TODO: [ExpectedException(typeof(NotSupportedException))]
    public void TestLoadMessageMimeVersionInvalid()
    {
      MimeMessage.LoadMessage("MIME-Version: \"1.0 (comment)\"\r\n");
    }

    [Test]
    public void TestLoadMessageMimeVersionWithComment()
    {
      MimeMessage.LoadMessage("MIME-Version: 1.0 (comment)\r\n");
    }

    [Test]
    public void TestLoadMessageMimeVersionQuoted()
    {
      MimeMessage.LoadMessage("MIME-Version: \"1.0\"\r\n");
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestLoadMessageFromStringLargeMessage(LoadMessageMode mode)
    {
      var header = "Content-Type: text/plain\r\n\r\n";
      var body = new StringBuilder();

      var line = "0123456789"; // 10

      line += line; // 20
      line += line; // 40
      line += line; // 80

      while (body.Length <= 40960) {
        body.AppendLine(line);
      }

      LoadMessage(header + body, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(body.ToString(),
                        mime.ReadContentAsText());
      });
    }

    private class NonSeekableMemoryStream : MemoryStream {
      public NonSeekableMemoryStream(byte[] buffer)
        : base(buffer)
      {
      }

      public override bool CanSeek {
        get { return false; }
      }

      public override long Position {
        get { throw new NotSupportedException(); }
        set { throw new NotSupportedException(); }
      }

      public override long Length {
        get { throw new NotSupportedException(); }
      }

      public override void SetLength(long value)
      {
        throw new NotSupportedException();
      }

      public override long Seek(long offset, SeekOrigin loc)
      {
        throw new NotSupportedException();
      }
    }

    [Test]
    public void TestLoadMessageFromNonSeekableStream()
    {
      var stream = new NonSeekableMemoryStream(Charsets.Latin1.GetBytes(@"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n")));
      var mime = MimeMessage.Load(stream);

      Assert.AreEqual("multipart message body\r\n", mime.ReadContentAsText());

      Assert.AreEqual(2, mime.SubParts.Count);
      
      Assert.AreEqual("first part, text/plain\r\n", mime.SubParts[0].ReadContentAsText());
      
      Assert.AreEqual("<html><body><p>second part, text/html</p></body></html>",
                      mime.SubParts[1].ReadContentAsText());
    }

    private void UsingTestFile(string content, Action<FileInfo> testAction)
    {
      const string file = "test.eml";

      try {
        if (File.Exists(file))
          File.Delete(file);

        File.WriteAllText(file, content, Charsets.Latin1);

        testAction(new FileInfo(file));
      }
      finally {
        if (File.Exists(file))
          File.Delete(file);
      }
    }

    [Test]
    public void TestLoadMessageNonBufferedSinglePart()
    {
      var message = @"Content-Type: text/plain
Subject: single part

<content>".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          mime.ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream);

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream);

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream);

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<content>");

            Assert.AreEqual(offset, fs.Position);
            Assert.AreEqual("<content>".Length, stream.Length);
          });


          Assert.AreEqual("<content>",
                          mime.ReadContentAsText());
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedMultiPart()
    {
      var message = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          mime.ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream, "<content>");

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "<content>");

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream, "<content>");

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<content>");

            Assert.AreEqual(offset, fs.Position, "position of <content>");
            Assert.AreEqual("<content>".Length, stream.Length, "length of <content>");
          });

          Assert.AreEqual("<content>",
                          mime.ReadContentAsText(), "content of <content>");

          mime.SubParts[0].ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream, "<part#0>");

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "<part#0>");

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream, "<part#0>");

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<part#0>");

            Assert.AreEqual(offset, fs.Position, "position of <part#0>");
            Assert.AreEqual("<part#0>".Length, stream.Length, "length of<part#0>");
          });

          Assert.AreEqual("<part#0>",
                          mime.SubParts[0].ReadContentAsText(), "content of <part#0>");

          mime.SubParts[1].ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream, "<part#1>");

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "<part#1>");

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream, "<part#1>");

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<part#1>");

            Assert.AreEqual(offset, fs.Position, "position of <part#1>");
            Assert.AreEqual("<part#1>".Length, stream.Length, "length of <part#1>");
          });

          Assert.AreEqual("<part#1>",
                          mime.SubParts[1].ReadContentAsText(), "content of <part#1>");
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedMultiPartDecode()
    {
      var message = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Content-Transfer-Encoding: base64
Subject: multi part

PGNvbnRlbnQ+
--==boundary==
Content-Type: text/plain
Content-Transfer-Encoding: base64

PHBhcnQjMD4=
--==boundary==
Content-Type: text/plain
Content-Transfer-Encoding: quoted-printable

=3Cpart=231=3E
--==boundary==
Content-Type: text/plain
Content-Transfer-Encoding: x-uuencode

begin 664 test.txt
(/'!A<G0C,CX`
`
end
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          Assert.AreEqual("<content>",
                          mime.ReadContentAsText(), "content of <content>");
          Assert.AreEqual("<part#0>",
                          mime.SubParts[0].ReadContentAsText(), "content of <part#0>");
          Assert.AreEqual("<part#1>",
                          mime.SubParts[1].ReadContentAsText(), "content of <part#1>");
          Assert.AreEqual("<part#2>",
                          mime.SubParts[2].ReadContentAsText(), "content of <part#2>");
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedMultiPartNested()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          mime.ReadContent(delegate(Stream stream) {
            Assert.DoesNotThrow(delegate {
              var ncs = stream as NonClosingStream;
              var ps = ncs.InnerStream as PartialStream;
              var fs = ps.InnerStream as FileStream;
              
              Assert.IsNotNull(fs);
            });
          });

          Assert.AreEqual("<content>",
                          mime.ReadContentAsText(), "content of <content>");
          Assert.AreEqual("<part#0>",
                          mime.SubParts[0].ReadContentAsText(), "content of <part#0>");
          Assert.AreEqual("<nested>",
                          mime.SubParts[1].ReadContentAsText(), "content of <nested>");
          Assert.AreEqual("<nested#0>",
                          mime.SubParts[1].SubParts[0].ReadContentAsText(), "content of <nested#0>");
          Assert.AreEqual("<nested#1>",
                          mime.SubParts[1].SubParts[1].ReadContentAsText(), "content of <nested#1>");
          Assert.AreEqual("<part#1>",
                          mime.SubParts[2].ReadContentAsText(), "content of <part#1>");
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedExceptionThrownFromParser()
    {
      var message = @"Content-Type: text/plain; charset=x-foo
Subject: single part

<content>".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        var ex = Assert.Throws<NotSupportedException>(delegate {
          MimeMessage.Load(file.FullName, false, delegate(string name) {
            throw new NotSupportedException("unsupported operation");
          });
        });

        Assert.AreEqual("unsupported operation", ex.Message);
      });
    }

    [Test]
    public void TestHeaders()
    {
      var message =
        "content-type: text/plain\r\n" +
          "FROM: from@example.com\r\n" + 
          "tO: to@example.com\r\n";

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("CONTENT-TYPE"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
      Assert.AreSame(mime.Headers["CONTENT-TYPE"], mime.Headers["Content-Type"]);

      Assert.IsTrue(mime.Headers.Contains("from"));
      Assert.AreEqual("from@example.com", mime.Headers["FROM"].Value);

      Assert.IsTrue(mime.Headers.Contains("to"));
      Assert.AreEqual("to@example.com", mime.Headers["TO"].Value);
    }

    [Test]
    public void TestHeadersMultipleHeader()
    {
      var message = @"MIME-Version: 1.0\r\n
Received: #1
Received: #2
Received: #3";

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsTrue(mime.Headers.Contains("Received"));

      Assert.AreEqual(3, mime.Headers.GetHeaders("Received").Count());
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#1").Count(), "#1");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#2").Count(), "#2");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#3").Count(), "#2");
    }

    [Test]
    public void TestParentPart()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsNull(mime.ParentPart, "root");
      Assert.AreSame(mime, mime.SubParts[0].ParentPart, "root.0");
      Assert.AreSame(mime, mime.SubParts[1].ParentPart, "root.1");
      Assert.AreSame(mime.SubParts[1], mime.SubParts[1].SubParts[0].ParentPart, "root.1.0");
      Assert.AreSame(mime.SubParts[1], mime.SubParts[1].SubParts[1].ParentPart, "root.1.1");
      Assert.AreSame(mime, mime.SubParts[2].ParentPart, "root.2");
    }

    [Test]
    public void TestRootPart()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreSame(mime, mime.RootPart, "root");
      Assert.AreSame(mime, mime.SubParts[0].RootPart, "root.0");
      Assert.AreSame(mime, mime.SubParts[1].RootPart, "root.1");
      Assert.AreSame(mime, mime.SubParts[1].SubParts[0].RootPart, "root.1.0");
      Assert.AreSame(mime, mime.SubParts[1].SubParts[1].RootPart, "root.1.1");
      Assert.AreSame(mime, mime.SubParts[2].RootPart, "root.2");
    }

    [Test]
    public void TestAllParts()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);
      var parts = mime.AllParts;

      Assert.AreEqual(5, parts.Length);
      Assert.AreSame(mime.SubParts[0], parts[0], "root.0");
      Assert.AreSame(mime.SubParts[1], parts[1], "root.1");
      Assert.AreSame(mime.SubParts[1].SubParts[0], parts[2], "root.1.0");
      Assert.AreSame(mime.SubParts[1].SubParts[1], parts[3], "root.1.1");
      Assert.AreSame(mime.SubParts[2], parts[4], "root.2");
    }

    [Test]
    public void TestPartSpecifier()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreEqual(string.Empty, mime.PartSpecifier);
      Assert.AreEqual("1", mime.SubParts[0].PartSpecifier, "root.0");
      Assert.AreEqual("2", mime.SubParts[1].PartSpecifier, "root.1");
      Assert.AreEqual("2.1", mime.SubParts[1].SubParts[0].PartSpecifier, "root.1.0");
      Assert.AreEqual("2.2", mime.SubParts[1].SubParts[1].PartSpecifier, "root.1.1");
      Assert.AreEqual("3", mime.SubParts[2].PartSpecifier, "root.2");
    }

    [Test]
    public void TestEncapsulatedMessage()
    {
      var message = @"Content-Type: message/rfc822

Return-Path: <root@xxx.xxxxxx.xxx>
Date: Fri, 21 Dec 2007 02:02:15 +0900
Content-Type: text/plain; charset=UTF-8
Auto-Submitted: auto-generated
X-Cron-Env: <MAILTO=root,postmaster,webmaster,clamav>
X-Cron-Env: <SHELL=/bin/sh>
X-Cron-Env: <HOME=/root>
X-Cron-Env: <PATH=/usr/bin:/bin>
X-Cron-Env: <LOGNAME=root>
X-Cron-Env: <USER=root>

ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Update failed. Your network may be down or none of the mirrors listed in freshclam.conf is working.

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx--
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreEqual(0, mime.SubParts.Count, "SubParts.Count");
      Assert.IsEmpty(mime.ReadContentAsText(), "ReacContentAsText");

      var encapsulated = mime.EncapsulatedMessage;

      Assert.IsNotNull(encapsulated, "EncapsulatedMessage");
      Assert.AreEqual(10, encapsulated.Headers.Count, "Headers.Count");
      Assert.AreEqual("Return-Path", encapsulated.Headers.First().Name, "Headers.First().Name");
      Assert.AreEqual("<root@xxx.xxxxxx.xxx>", encapsulated.Headers.First().Value, "Headers.First().Value");
      Assert.AreEqual("X-Cron-Env", encapsulated.Headers.Last().Name, "Headers.Last().Name");
      Assert.AreEqual("<USER=root>", encapsulated.Headers.Last().Value, "Headers.Last().Value");
      StringAssert.StartsWith("ERROR: Can't query current.cvd.clamav.net", encapsulated.ReadContentAsText());
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestDispose(LoadMessageMode mode)
    {
      var content = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var part1 = mime.SubParts[0];
        var part2 = mime.SubParts[1];

        mime.Dispose();

        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.Headers));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.SubParts));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.ParentPart));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.RootPart));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.AllParts));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.PartSpecifier));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.EncapsulatedMessage));
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContent(delegate(StreamReader r) {
          r.ReadLine();
        }));
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContent(delegate(BinaryReader r) {
          r.ReadByte();
        }));
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContentAsText());
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContentAsBinary());
        Assert.Throws<ObjectDisposedException>(() => mime.CopyContentTo(Stream.Null));
        Assert.Throws<ObjectDisposedException>(() => mime.SaveContent("content.tmp"));
        Assert.Throws<ObjectDisposedException>(() => mime.Save("content.tmp"));
        Assert.Throws<ObjectDisposedException>(() => mime.Save(Stream.Null));
        Assert.Throws<ObjectDisposedException>(() => mime.ToString());

        Assert.Throws<ObjectDisposedException>(() => part1.ReadContentAsText());
        Assert.Throws<ObjectDisposedException>(() => part2.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.Dispose(), "re-Dispose");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestDisposeSubPart(LoadMessageMode mode)
    {
      var content = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var part1 = mime.SubParts[0];
        var part2 = mime.SubParts[1];

        part1.Dispose();

        Assert.Throws<ObjectDisposedException>(() => part1.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.ReadContentAsText());
        Assert.DoesNotThrow(() => part2.ReadContentAsText());

        part2.Dispose();

        Assert.Throws<ObjectDisposedException>(() => part2.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.Dispose(), "Dispose root");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestDisposeEncapsulatedMessage(LoadMessageMode mode)
    {
      var content = @"Content-Type: message/delivery-status

Reporting-MTA: dns; xxx.xxxxxx.xxx
Arrival-Date: Fri, 21 Dec 2007 02:06:24 +0900

Final-Recipient: RFC822; root@xxx.xxxxxx.xxx
X-Actual-Recipient: RFC822; xxxxxx@xxx.xxxxxx.xxx
Action: delayed
Status: 4.4.3
Last-Attempt-Date: Fri, 21 Dec 2007 07:04:35 +0900
Will-Retry-Until: Wed, 26 Dec 2007 02:06:24 +0900";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var encapsulated = mime.EncapsulatedMessage;

        encapsulated.Dispose();

        Assert.Throws<ObjectDisposedException>(() => encapsulated.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.Dispose(), "Dispose root");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentAsText(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual("content", mime.ReadContentAsText());
        Assert.AreEqual("content", mime.ReadContentAsText(), "read again");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentAsBinary(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"), mime.ReadContentAsBinary());
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"), mime.ReadContentAsBinary(), "read again");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTypeNotExist(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Subject: transfer encoding not exist

decode as text/plain; charset=us-ascii";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual("decode as text/plain; charset=us-ascii", mime.ReadContentAsText());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTypeNonText(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: application/octet-stream
Subject: transfer encoding not exist

decode as binary stream";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var text = Encoding.ASCII.GetBytes("decode as binary stream");

        Assert.AreEqual(MimeType.ApplicationOctetStream, mime.MimeType);
        Assert.AreEqual(text, mime.ReadContentAsBinary());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncoding7BitMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: 7bit transfer encoding

7bit transfer encoding";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Charsets.ASCII, mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.SevenBit, mime.TransferEncoding);
        Assert.AreEqual("7bit", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("7bit transfer encoding", mime.ReadContentAsText());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncoding8BitMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=iso-8859-1
Content-Transfer-Encoding: 8bit
Subject: 8bit transfer encoding

" + "8bit transfer encoding \u0080 \u00cf \u00ff\r\n";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Charsets.Latin1, mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.EightBit, mime.TransferEncoding);
        Assert.AreEqual("8bit", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("8bit transfer encoding \u0080 \u00cf \u00ff\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingBinaryMessage(LoadMessageMode mode)
    {
      var message = new MemoryStream(0x200);

      var header = Charsets.Latin1.GetBytes(@"MIME-Version: 1.0
Content-Type: application/octet-stream
Content-Transfer-Encoding: binary
Subject: binary transfer encoding

");

      message.Write(header, 0, header.Length);

      var bodyStream = new MemoryStream(); 

      for (var i = 0; i < 0x100; i++) {
        bodyStream.WriteByte((byte)i);
      }

      var body = bodyStream.ToArray();

      message.Write(body, 0, body.Length);

      var content = Charsets.Latin1.GetString(message.ToArray());

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.ApplicationOctetStream, mime.MimeType);
        Assert.AreEqual(ContentTransferEncodingMethod.Binary, mime.TransferEncoding);
        Assert.AreEqual("binary", mime.Headers["Content-Transfer-Encoding"].Value);

        Assert.AreEqual(body, mime.ReadContentAsBinary());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingBase64Message(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=shift_jis
Content-Transfer-Encoding: base64
Subject: base64 transfer encoding

ir+OmmFiY4KpgsgxMjODSoNpDQqCUIJRglKCgYKCgoMNCg==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.Base64, mime.TransferEncoding);
        Assert.AreEqual("base64", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingQuotedPrintableMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: quoted-printable
Subject: quoted printable transfer encoding

=B4=C1=BB=FAabc=A4=AB=A4=CA123=A5=AB=A5=CA=0D=0A=A3=B1=A3=B2=A3=B3=A3=E1=A3=
=E2=A3=E3=0D=0A";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.QuotedPrintable, mime.TransferEncoding);
        Assert.AreEqual("quoted-printable", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingUUEncodeMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: uuencode
Subject: uuencode transfer encoding

begin 664 test.txt
BM,&[^F%B8Z2KI,HQ,C.EJZ7*#0JCL:.RH[.CX:/BH^,-""@``
`
end

begin 664 extra.txt
+97AT<F$@9FEL90H`
`
end
";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.UUEncode, mime.TransferEncoding);
        Assert.AreEqual("uuencode", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingUnsupported(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: unsupported
Subject: unsupported transfer encoding

unsupported transfer encoding";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.Unknown, mime.TransferEncoding);
        Assert.AreEqual("unsupported", mime.Headers["Content-Transfer-Encoding"].Value);

        try {
          mime.ReadContentAsText();

          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStream(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(false, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("Y29udGVudA==", Encoding.ASCII.GetString(stream.ToArray()));
          }
        });

        mime.ReadContent(false, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("Y29udGVudA==", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamDecode(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(true, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()));
          }
        });

        mime.ReadContent(true, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamCloseStreamPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(false, delegate(Stream contentStream) {
          contentStream.Close();
        });

        mime.ReadContent(false, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamCloseStreamNonPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(true, delegate(Stream contentStream) {
          contentStream.Close();
        });

        mime.ReadContent(true, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamReader(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual(Charsets.Latin1, reader.CurrentEncoding);
          Assert.AreEqual("content", reader.ReadToEnd());
        });

        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual("content", reader.ReadToEnd(), "read again");
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentBinaryReader(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7));
          Assert.AreEqual(-1, reader.PeekChar());
        });

        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7),
                          "read again");
          Assert.AreEqual(-1, reader.PeekChar());
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamReaderCloseReaderPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(StreamReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual("content", reader.ReadToEnd(), "read again");
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamReaderCloseReaderNonPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(StreamReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual("content", reader.ReadToEnd(), "read again");
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentBinaryReaderCloseReaderPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(BinaryReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7),
                          "read again");
          Assert.AreEqual(-1, reader.PeekChar());
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentBinaryReaderCloseReaderNonPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(BinaryReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7),
                          "read again");
          Assert.AreEqual(-1, reader.PeekChar());
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestCopyContentTo(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        using (var outStream = new MemoryStream()) {
          mime.CopyContentTo(outStream, false);

          Assert.AreEqual(Encoding.ASCII.GetBytes("Y29udGVudA=="),
                          outStream.ToArray());
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestCopyContentToDecodeContent(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        using (var outStream = new MemoryStream()) {
          mime.CopyContentTo(outStream, true);

          Assert.AreEqual(Encoding.ASCII.GetBytes("content"),
                          outStream.ToArray());
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestSaveContent(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        const string file = "content.txt";

        try {
          if (File.Exists(file))
            File.Delete(file);

          mime.SaveContent(file, false);

          Assert.IsTrue(File.Exists(file));
          Assert.AreEqual("Y29udGVudA==",
                          File.ReadAllText(file, Encoding.ASCII));
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestSaveContentOverwriteExistingFile(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        const string file = "content.txt";

        try {
          if (File.Exists(file))
            File.Delete(file);

          File.WriteAllText(file, new string('*', 1024));

          mime.SaveContent(file, false);

          Assert.IsTrue(File.Exists(file));
          Assert.AreEqual("Y29udGVudA==",
                          File.ReadAllText(file, Encoding.ASCII));
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestSaveContentDecodeContent(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        const string file = "content.txt";

        try {
          if (File.Exists(file))
            File.Delete(file);

          mime.SaveContent(file, true);

          Assert.IsTrue(File.Exists(file));
          Assert.AreEqual("content",
                          File.ReadAllText(file, Encoding.ASCII));
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      });
    }

    private void TestReassembledMessage(MimeMessage expected, MimeMessage reassembled)
    {
      Assert.IsNotNull(reassembled, "reassembled message");

      Assert.AreEqual(expected.MimeType, reassembled.MimeType, "mime type");
      Assert.AreEqual(expected.TransferEncoding, reassembled.TransferEncoding, "transfer encoding");

      Assert.AreEqual(expected.ReadContentAsText(),
                      reassembled.ReadContentAsText(), "content");

      Assert.AreEqual(expected.Headers.Count, reassembled.Headers.Count, "count of header fields");

      foreach (var expectedHeader in expected.Headers) {
        Assert.IsTrue(reassembled.Headers.Contains(expectedHeader.Name), "contains " + expectedHeader.Name);
        Assert.AreEqual(expectedHeader.Value,
                        reassembled.Headers[expectedHeader.Name].Value, "value of " + expectedHeader.Name);
      }
    }

    [TestCase("class+params")]
    [TestCase("class+IEnumerable")]
    [TestCase("instance")]
    public void TestReassembleFragmentedMessage1(string testMethod)
    {
      var partialFirst = MimeMessage.LoadMessage(@"X-Weird-Header-1: Foo
From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Subject: Audio mail (part 1 of 2)
Message-ID: <id1@host.com>
MIME-Version: 1.0
Content-type: message/partial; id=""ABC@host.com"";
              number=1; total=2

X-Weird-Header-1: Bar
X-Weird-Header-2: Hello
Message-ID: <anotherid@foo.com>
Subject: Audio mail
MIME-Version: 1.0
Content-type: audio/basic
Content-transfer-encoding: base64

Li4uIGZpcnN0IGhhbGYgb2YgZW5jb2RlZCBhdWRpbyBkYXRhIGdvZXMgaGVyZSAuLi4=
"
      );
      var partialSecond = MimeMessage.LoadMessage(@"From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Subject: Audio mail (part 2 of 2)
MIME-Version: 1.0
Message-ID: <id2@host.com>
Content-type: message/partial;
              id=""ABC@host.com""; number=2; total=2

Li4uIHNlY29uZCBoYWxmIG9mIGVuY29kZWQgYXVkaW8gZGF0YSBnb2VzIGhlcmUgLi4u
"
      );
      var expected = MimeMessage.LoadMessage(@"X-Weird-Header-1: Foo
From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Subject: Audio mail
Message-ID: <anotherid@foo.com>
MIME-Version: 1.0
Content-type: audio/basic
Content-transfer-encoding: base64

Li4uIGZpcnN0IGhhbGYgb2YgZW5jb2RlZCBhdWRpbyBkYXRhIGdvZXMgaGVyZSAuLi4=
Li4uIHNlY29uZCBoYWxmIG9mIGVuY29kZWQgYXVkaW8gZGF0YSBnb2VzIGhlcmUgLi4u
"
      );

      MimeMessage reassembled = null;

      if (testMethod == "class+params")
        reassembled = MimeMessage.ReassembleFragmentedMessage(partialFirst, partialSecond);
      else if (testMethod == "class+IEnumerable")
        reassembled = MimeMessage.ReassembleFragmentedMessage(new[] {partialFirst, partialSecond});
      else if (testMethod == "instance")
        reassembled = partialFirst.ReassembleFragmentedMessage(partialSecond);
      else
        Assert.Fail("invalid test method");

      TestReassembledMessage(expected, reassembled);

      Assert.AreNotSame(partialFirst.Headers["X-Weird-Header-1"],
                        reassembled.Headers["X-Weird-Header-1"],
                        "instance of X-Weird-Header-1");
      Assert.AreNotSame(partialFirst.Headers["Date"],
                        reassembled.Headers["Date"],
                        "instance of Date");
    }

    [TestCase("class+params")]
    [TestCase("class+IEnumerable")]
    [TestCase("instance")]
    public void TestReassembleFragmentedMessage2(string testMethod)
    {
      var partialFirst = MimeMessage.LoadMessage(@"Content-Type: message/partial; number=1
X-Reassembled-Header: foo

CONTENT-TYPE: text/plain
content-transfer-encoding: 7bit
X-Ignored-Header: bar

partial#1
");
      var partialSecond = MimeMessage.LoadMessage(@"Content-Type: MESSAGE/PARTIAL; number=2

partial#2
");
      var partialThird = MimeMessage.LoadMessage(@"Content-Type: message/partial; NUMBER=3
X-Ignored-Header: baz

partial#3
");
      var expected = MimeMessage.LoadMessage(@"CONTENT-TYPE: text/plain
content-transfer-encoding: 7bit
X-Reassembled-Header: foo

partial#1
partial#2
partial#3
");

      MimeMessage reassembled = null;

      if (testMethod == "class+params")
        reassembled = MimeMessage.ReassembleFragmentedMessage(partialThird, partialFirst, new[] {partialSecond});
      else if (testMethod == "class+IEnumerable")
        reassembled = MimeMessage.ReassembleFragmentedMessage(new[] {partialThird, partialFirst, partialSecond});
      else if (testMethod == "instance")
        reassembled = partialThird.ReassembleFragmentedMessage(partialFirst, new[] {partialSecond});
      else
        Assert.Fail("invalid test method");

      TestReassembledMessage(expected, reassembled);

      Assert.AreNotSame(partialFirst.Headers["X-Reassembled-Header"],
                        reassembled.Headers["X-Reassembled-Header"],
                        "instance of X-Reassembled-Header");
    }

    [Test]
    public void TestReassembleFragmentedMessageTotalSpecified()
    {
      var fragmentedMessages = (new string[] {
        @"Content-Type: text/plain; number=3

this message must be ignored
",
        @"Content-Type: message/partial; number=2; total=2

partial#2
",
        @"Content-Type: message/partial; number=1

Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
",
        @"Content-Type: message/partial; number=4

this fragment must be ignored
",
        @"Content-Type: message/partial; number=3; total=3

partial#3
",
      }).Select<string, MimeMessage>(MimeMessage.LoadMessage);

      var expected = MimeMessage.LoadMessage(@"Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
partial#2
partial#3
");

      TestReassembledMessage(expected,
                             MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)fragmentedMessages));
    }

    [Test]
    public void TestReassembleFragmentedMessageIDSpecified()
    {
      var firstFragmentedMessage = MimeMessage.LoadMessage(@"Content-Type: message/partial; number=2; id=""foo""

partial#2
");
      var fragmentedMessages = (new string[] {
        @"Content-Type: text/plain; number=3; id=""foo""

this message must be ignored
",
        @"Content-Type: message/partial; number=1; id=""bar""

Content-Type: text/plain
Content-Transfer-Encoding: 7bit

this fragment must be ignored
",
        @"Content-Type: message/partial; number=3; id=""foo""

partial#3
",
        @"Content-Type: message/partial; number=1; id=""foo""

Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
",
      }).Select<string, MimeMessage>(MimeMessage.LoadMessage).ToArray();

      var expected = MimeMessage.LoadMessage(@"Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
partial#2
partial#3
");

      TestReassembledMessage(expected,
                             firstFragmentedMessage.ReassembleFragmentedMessage(fragmentedMessages[0], fragmentedMessages.Slice(1)));
    }

    [Test]
    public void TestReassembleFragmentedMessageIgnoreInvalidNumber()
    {
      var fragmentedMessages = (new string[] {
        @"Content-Type: message/partial; number=1

Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
",
        @"Content-Type: message/partial; number=0

this fragment must be ignored
",
        @"Content-Type: message/partial; number=-1

this fragment must be ignored
",
        @"Content-Type: message/partial; number=2

partial#2
",
      }).Select<string, MimeMessage>(MimeMessage.LoadMessage);

      var expected = MimeMessage.LoadMessage(@"Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
partial#2
");

      TestReassembledMessage(expected,
                             MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)fragmentedMessages));
    }

    [Test]
    public void TestReassembleFragmentedMessageInstanceIsNotFragmentedMessage()
    {
      var nonPartial = MimeMessage.LoadMessage(@"Content-Type: text/plain

text
");
      var other = MimeMessage.LoadMessage(@"Content-Type: message/partial; number=2

partial#2
");

      Assert.Throws<InvalidOperationException>(() => nonPartial.ReassembleFragmentedMessage(other));
    }

    [Test]
    public void TestReassembleFragmentedMessageArgumentNull()
    {
      var partialMessage = MimeMessage.LoadMessage(@"Content-Type: message/partial; number=1

partial#1");

      Assert.AreEqual("fragmentedMessage1",
                      Assert.Throws<ArgumentNullException>(() => MimeMessage.ReassembleFragmentedMessage((MimeMessage)null, (MimeMessage)null)).ParamName);
      Assert.AreEqual("fragmentedMessage1",
                      Assert.Throws<ArgumentNullException>(() => MimeMessage.ReassembleFragmentedMessage((MimeMessage)null, partialMessage)).ParamName);
      Assert.AreEqual("fragmentedMessage2",
                      Assert.Throws<ArgumentNullException>(() => MimeMessage.ReassembleFragmentedMessage(partialMessage, (MimeMessage)null)).ParamName);
      Assert.AreEqual("fragmentedMessages",
                      Assert.Throws<ArgumentNullException>(() => MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)null)).ParamName);
      Assert.AreEqual("fragmentedMessage",
                      Assert.Throws<ArgumentNullException>(() => partialMessage.ReassembleFragmentedMessage((MimeMessage)null)).ParamName);
    }

    [Test]
    public void TestReassembleFragmentedMessageArgumentMissingEnclosingMessage()
    {
      var fragmentedMessages = (new string[] {
        @"Content-Type: message/partial; number=2

partial#2
",
        @"Content-Type: message/partial; number=3

partial#3
",
      }).Select<string, MimeMessage>(MimeMessage.LoadMessage);

      var ex = Assert.Throws<ArgumentException>(() => MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)fragmentedMessages));

      Assert.AreEqual("fragmentedMessages", ex.ParamName);
      StringAssert.Contains("enclosing message", ex.Message);
    }

    [Test]
    public void TestReassembleFragmentedMessageArgumentMissingFragmentTotalNotSpecified()
    {
      var fragmentedMessages = (new string[] {
        @"Content-Type: message/partial; number=1

Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
",
        @"Content-Type: message/partial; number=2

partial#2
",
        @"Content-Type: message/partial; number=3

partial#3
",
        @"Content-Type: message/partial; number=4

partial#4
",
      }).Select<string, MimeMessage>(MimeMessage.LoadMessage);

      for (var number = 2; number <= 4; number++) {
        var messages = fragmentedMessages.Where((f, n) => n != number - 1);

        if (number == 4) {
          Assert.DoesNotThrow(() => MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)messages));
        }
        else {
          var ex = Assert.Throws<ArgumentException>(() => MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)messages));

          Assert.AreEqual("fragmentedMessages", ex.ParamName);
          StringAssert.Contains("number=" + number.ToString(), ex.Message);
        }
      }
    }

    [Test]
    public void TestReassembleFragmentedMessageArgumentMissingFragmentTotalSpecified()
    {
      var fragmentedMessages = (new string[] {
        @"Content-Type: message/partial; number=1; total=4

Content-Type: text/plain
Content-Transfer-Encoding: 7bit

partial#1
",
        @"Content-Type: message/partial; number=2

partial#2
",
        @"Content-Type: message/partial; number=3

partial#3
",
        @"Content-Type: message/partial; number=4

partial#4
",
      }).Select<string, MimeMessage>(MimeMessage.LoadMessage);

      for (var number = 2; number <= 4; number++) {
        var messages = fragmentedMessages.Where((f, n) => n != number - 1);

        var ex = Assert.Throws<ArgumentException>(() => MimeMessage.ReassembleFragmentedMessage((IEnumerable<IMimeMessage>)messages));

        Assert.AreEqual("fragmentedMessages", ex.ParamName);
        StringAssert.Contains("number=" + number.ToString(), ex.Message);
      }
    }

    [Test]
    public void TestSave()
    {
      var expected = @"Content-Type: text/plain

content".Replace("\r\n", "\n").Replace("\n", "\r\n");

      const string file = "test.eml";

      try {
        if (File.Exists(file))
          File.Delete(file);

        MimeMessage.LoadMessage(expected).Save(file);

        Assert.AreEqual(expected,
                        File.ReadAllText(file, Charsets.Latin1));
      }
      finally {
        if (File.Exists(file))
          File.Delete(file);
      }
    }

    [Test]
    public void TestSaveOverwriteExistingFile()
    {
      var expected = @"Content-Type: text/plain

content".Replace("\r\n", "\n").Replace("\n", "\r\n");

      const string file = "test.eml";

      try {
        if (File.Exists(file))
          File.Delete(file);

        File.WriteAllText(file, new string('*', 1024));

        MimeMessage.LoadMessage(expected).Save(file);

        Assert.AreEqual(expected,
                        File.ReadAllText(file, Charsets.Latin1));
      }
      finally {
        if (File.Exists(file))
          File.Delete(file);
      }
    }

    private void TestSave(string expectedContent, Action<string> testSaveAction)
    {
      const string file = "test.eml";

      try {
        if (File.Exists(file))
          File.Delete(file);

        testSaveAction(file);

        Assert.AreEqual(expectedContent,
                        File.ReadAllText(file, Charsets.Latin1));
      }
      finally {
        if (File.Exists(file))
          File.Delete(file);
      }
    }

    [Test]
    public void TestSaveDiscrete()
    {
      var expected = @"From: John Doe <jdoe@machine.example>
To: Mary Smith <mary@example.net>
Subject: Saying Hello
Date: Fri, 21 Nov 1997 09:55:06 -0600
Message-ID: <1234@local.machine.example>

This is a message just to say hello.
So, ""Hello"".
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        MimeMessage.LoadMessage(expected).Save(filename);
      });
    }

    [Test]
    public void TestSaveComposite()
    {
      var expected = @"From: Nathaniel Borenstein <nsb@bellcore.com>
To: Ned Freed <ned@innosoft.com>
Date: Sun, 21 Mar 1993 23:56:48 -0800 (PST)
Subject: Sample message
MIME-Version: 1.0
Content-type: multipart/mixed; boundary=""simple boundary""

This is the preamble.  It is to be ignored, though it
is a handy place for composition agents to include an
explanatory note to non-MIME conformant readers.

--simple boundary

This is implicitly typed plain US-ASCII text.
It does NOT end with a linebreak.
--simple boundary
Content-type: text/plain; charset=us-ascii

This is explicitly typed plain US-ASCII text.
It DOES end with a linebreak.

--simple boundary--
".Replace("\r\n", "\n").Replace("\n", "\r\n");
      var content = expected + "\r\nThis is the epilogue.  It is also to be ignored.";

      TestSave(expected, delegate(string filename) {
        MimeMessage.LoadMessage(content).Save(filename);
      });
    }

    [Test]
    public void TestSaveMessageCapsulated()
    {
      var expected = @"Content-type: message/external-body;
              access-type=local-file;
              name=""/u/nsb/Me.jpeg""

Content-type: image/jpeg
Content-ID: <id42@guppylake.bellcore.com>
Content-Transfer-Encoding: binary

THIS IS NOT REALLY THE BODY!".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        Console.WriteLine(MimeMessage.LoadMessage(expected).ToString());
        MimeMessage.LoadMessage(expected).Save(filename);
      });
    }

    [Test]
    public void TestSaveNested()
    {
      var expected = @"From: Moderator-Address
To: Recipient-List
Date: Mon, 22 Mar 1994 13:34:51 +0000
Subject: Internet Digest, volume 42
MIME-Version: 1.0
Content-Type: multipart/mixed;
              boundary=""---- main boundary ----""

------ main boundary ----

...Introductory text or table of contents...

------ main boundary ----
Content-Type: multipart/digest;
              boundary=""---- next message ----""

------ next message ----

From: someone-else
Date: Fri, 26 Mar 1993 11:13:32 +0200
Subject: my opinion

...body goes here ...

------ next message ----

From: someone-else-again
Date: Fri, 26 Mar 1993 10:07:13 -0500
Subject: my different opinion

... another body goes here ...

------ next message ------

------ main boundary ------
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        MimeMessage.LoadMessage(expected).Save(filename);
      });
    }

    [Test]
    public void TestSavePartialMessage()
    {
      var expected = @"From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Subject: Audio mail (part 2 of 2)
MIME-Version: 1.0
Message-ID: <id2@host.com>
Content-type: message/partial;
              id=""ABC@host.com""; number=2; total=2

... second half of encoded audio data goes here ...
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        MimeMessage.LoadMessage(expected).Save(filename);
      });
    }

    [Test]
    public void TestSaveReassembledFragmentMessage()
    {
      var partialFirst = MimeMessage.LoadMessage(@"X-Weird-Header-1: Foo
From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Subject: Audio mail (part 1 of 2)
Message-ID: <id1@host.com>
MIME-Version: 1.0
Content-type: message/partial; id=""ABC@host.com"";
              number=1; total=2

X-Weird-Header-1: Bar
X-Weird-Header-2: Hello
Message-ID: <anotherid@foo.com>
Subject: Audio mail
MIME-Version: 1.0
Content-type: audio/basic
Content-transfer-encoding: base64

Li4uIGZpcnN0IGhhbGYgb2YgZW5jb2RlZCBhdWRpbyBkYXRhIGdvZXMgaGVyZSAuLi4=
".Replace("\r\n", "\n").Replace("\n", "\r\n"));
      var partialSecond = MimeMessage.LoadMessage(@"From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Subject: Audio mail (part 2 of 2)
MIME-Version: 1.0
Message-ID: <id2@host.com>
Content-type: message/partial;
              id=""ABC@host.com""; number=2; total=2

Li4uIHNlY29uZCBoYWxmIG9mIGVuY29kZWQgYXVkaW8gZGF0YSBnb2VzIGhlcmUgLi4u
".Replace("\r\n", "\n").Replace("\n", "\r\n"));
      var expected = @"X-Weird-Header-1: Foo
From: Bill@host.com
To: joe@otherhost.com
Date: Fri, 26 Mar 1993 12:59:38 -0500 (EST)
Message-ID: <anotherid@foo.com>
Subject: Audio mail
MIME-Version: 1.0
Content-type: audio/basic
Content-transfer-encoding: base64

Li4uIGZpcnN0IGhhbGYgb2YgZW5jb2RlZCBhdWRpbyBkYXRhIGdvZXMgaGVyZSAuLi4=
Li4uIHNlY29uZCBoYWxmIG9mIGVuY29kZWQgYXVkaW8gZGF0YSBnb2VzIGhlcmUgLi4u
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        MimeMessage.ReassembleFragmentedMessage(partialFirst, partialSecond).Save(filename);
      });
    }

    [Test]
    public void TestSaveHeaderEncoded()
    {
      var expected = @"MIME-Version: 1.0
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=

body".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        MimeMessage.LoadMessage(expected).Save(filename);
      });
    }

    [Test]
    public void TestSaveHeaderWhitespaces()
    {
      var expected = @"From: Pete(A nice \) chap) <pete(his account)@silly.test(his host)>
To:A Group(Some people)
     :Chris Jones <c@(Chris's host.)public.example>,
         joe@example.org,
  John <jdoe@one.test> (my dear friend); (the end of the group)
Cc:(Empty list)(start)Hidden recipients  :(nobody(that I know))  ;
Date: Thu,
      13
        Feb
          1969
      23:32
               -0330 (Newfoundland Time)
Message-ID:              <testabcd.1234@silly.test>

Testing.".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestSave(expected, delegate(string filename) {
        MimeMessage.LoadMessage(expected).Save(filename);
      });
    }

    [Test]
    public void TestToString()
    {
      var expected = @"Content-Type: text/plain

content".Replace("\r\n", "\n").Replace("\n", "\r\n");

      Assert.AreEqual(expected,
                      MimeMessage.LoadMessage(expected).ToString());
    }
  }
}