using System;
using System.IO;
using System.Linq;
using System.Text;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MimeMessageParsingTests : MimeMessageTestsBase {
    [Test]
    public void TestParseHeaderContentTypeMimeTypeInvalid1()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text; charset= iso-2022-jp

content");

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
    }

    [Test]
    public void TestParseHeaderContentTypeMimeTypeInvalid2()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: ; charset= iso-2022-jp

content");

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
    }

    [Test]
    public void TestParseHeaderContentTypeWithEmptyParameter()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: image/jpeg;

content");
      
      Assert.AreEqual(MimeType.CreateImageType("jpeg"), mime.MimeType);
    }

    [Test]
    public void TestParseHeaderContentDisposition()
    {
      var mime1 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Disposition: inline

content");

      Assert.AreEqual(MimeMessageDisposition.Inline, mime1.Disposition);

      var mime2 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Disposition: INLINE

content");

      Assert.AreEqual(MimeMessageDisposition.Inline, mime2.Disposition);

      var mime3 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Disposition: ATTACHMENT; filename=foo.txt

content");

      Assert.AreEqual(MimeMessageDisposition.Attachment, mime3.Disposition);
    }

    [Test]
    public void TestParseHeaderContentDispositionWithEmptyParameter()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Disposition: attachment;

content");

      Assert.AreEqual(MimeMessageDisposition.Attachment, mime.Disposition);
    }

    [Test]
    public void TestParseHeaderContentDispositionHasNoDisposition()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0

content");

      Assert.AreEqual(MimeMessageDisposition.None, mime.Disposition);
    }

    [Test]
    public void TestParseHeaderContentDispositionUnsupportedDisposition()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Disposition: x-foo

content");

      Assert.AreEqual(MimeMessageDisposition.Unknown, mime.Disposition);
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetContainsLeadingWhitespaces()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset= iso-2022-jp
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetContainsTrailingWhitespaces()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=iso-2022-jp 
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetQuoted()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=""iso-2022-jp""
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetQuotedContainsWhitespaces1()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset="" iso-2022-jp ""
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetQuotedContainsWhitespaces2()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=""shift jis""
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetIllegal1()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=shift-jis
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetIllegal2()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=euc_jp
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetIllegal3()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=iso_2022-jp
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetNotFoundOrUnsupported()
    {
      try {
        MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body");

        Assert.Fail("EncodingNotSupportedException not thrown");
      }
      catch (EncodingNotSupportedException) {
      }
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetUseFallbackCharset()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body", Encoding.UTF8);

      Assert.IsNotNull(mime.Charset);
      Assert.AreEqual(Encoding.UTF8, mime.Charset);
      Assert.AreEqual("body", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetSelectFallbackCharset()
    {
      var called = false;
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body", delegate(string name) {
        called = true;

        Assert.AreEqual("x-unsupported", name, "callback name arg");

        return Encoding.UTF8;
      });

      Assert.IsTrue(called);
      Assert.IsNotNull(mime.Charset);
      Assert.AreEqual(Encoding.UTF8, mime.Charset);
      Assert.AreEqual("body", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetSelectFallbackCharsetReturnNull()
    {
      var called = false;

      try {
        MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body", delegate(string name) {
          called = true;
          return null;
        }
        );

        Assert.Fail("EncodingNotSupportedException not thrown");
      }
      catch (EncodingNotSupportedException ex) {
        Assert.AreEqual("x-unsupported", ex.EncodingName, "EncodingName");
      }

      Assert.IsTrue(called);
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetSelectFallbackCharsetWithLoadFromFile()
    {
      const string testFile = "test.eml";

      try {
        if (File.Exists(testFile))
          File.Delete(testFile);

        File.WriteAllText(testFile, @"Content-Type: text/plain; charset=x-unsupported

body", Encoding.ASCII);

        using (var mime = MimeMessage.Load(testFile, delegate(string name) {
          Assert.AreEqual("x-unsupported", name, "callback name arg");

          return Encoding.UTF8;
        })) {
          Assert.IsNotNull(mime.Charset);
          Assert.AreEqual(Encoding.UTF8, mime.Charset);
          Assert.AreEqual("body", mime.ReadContentAsText());
        }
      }
      finally {
        if (File.Exists(testFile))
          File.Delete(testFile);
      }
    }

    [Test]
    public void TestParseBodyNewlineDelimitedByInternalStreamBuffer()
    {
      const int bufferSize = 1024;

      var message = "MIME-Version: 1.0\r\nSubject: ";
      var splittedAt = bufferSize - message.Length - 1;
      var expectedSubject = new string('x', splittedAt) + "\r\n\tyyy";

      message += new string('x', splittedAt);
      message += "\r";

      if (message.Length != bufferSize)
        Assert.Fail("not splitted");

      message += "\n";
      message += "\tyyy\r\n";
      message += "From: from@example.com\r\n\r\n";
      message += "body";

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual(expectedSubject, mime.Headers["Subject"].Value);

      Assert.IsTrue(mime.Headers.Contains("From"));
      Assert.AreEqual("from@example.com", mime.Headers["From"].Value);

      Assert.AreEqual("body", mime.ReadContentAsText());
    }

    
    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestParseBodyDiscrete(string newline, LoadMessageMode mode)
    {
      var content = @"Content-Type: text/plain

first line
second line".Replace("\r\n", "\n").Replace("\n", newline);
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
        
        Assert.AreEqual("first line" + newline + "second line",
                        mime.ReadContentAsText());
      });
    }
    
    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestParseBodyDiscreteEndsWithNewLine(string newline, LoadMessageMode mode)
    {
      var content = @"Content-Type: text/plain

first line
second line
".Replace("\r\n", "\n").Replace("\n", newline);
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        
        Assert.AreEqual("first line" + newline + "second line" + newline,
                        mime.ReadContentAsText());
      });
    }
    
    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestParseBodyDiscreteEndsWithEmptyLine(string newline, LoadMessageMode mode)
    {
      var content = @"Content-Type: text/plain

first line
second line

".Replace("\r\n", "\n").Replace("\n", newline);
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        
        Assert.AreEqual("first line" + newline + "second line" + newline + newline,
                        mime.ReadContentAsText());
      });
    }
    
    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestParseBodyComposite1(string newline, LoadMessageMode mode)
    {
      var content = @"From: Nathaniel Borenstein <nsb@bellcore.com>
To: Ned Freed <ned@innosoft.com>
Date: Sun, 21 Mar 1993 23:56:48 -0800 (PST)
Subject: Sample message
MIME-Version: 1.0
Content-type: multipart/mixed; boundary=""simple boundary""

This is the preamble.  It is to be ignored, though it
is a handy place for composition agents to include an
explanatory note to non-MIME conformant readers.

--simple boundary

This is implicitly typed plain US-ASCII text.
It does NOT end with a linebreak.
--simple boundary
Content-type: text/plain; charset=us-ascii

This is explicitly typed plain US-ASCII text.
It DOES end with a linebreak.

--simple boundary--

This is the epilogue.  It is also to be ignored.".Replace("\r\n", "\n").Replace("\n", newline);
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.MultipartMixed, mime.MimeType);
        Assert.AreEqual("simple boundary", mime.Boundary);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("multipart/mixed", mime.Headers["Content-Type"].GetValueWithoutParameter());
        Assert.AreEqual("simple boundary", mime.Headers["Content-Type"].GetParameter("boundary"));
        
        Assert.AreEqual("This is the preamble.  It is to be ignored, though it" + newline + 
                        "is a handy place for composition agents to include an" + newline + 
                        "explanatory note to non-MIME conformant readers." + newline,
                        mime.ReadContentAsText());
        
        Assert.AreEqual(2, mime.SubParts.Count);
        
        Assert.IsFalse(mime.SubParts[0].Headers.Contains("Content-Type"));
        Assert.AreEqual(MimeType.TextPlain, mime.SubParts[0].MimeType);
        Assert.AreEqual(Charsets.Latin1, mime.SubParts[0].Charset);
        Assert.AreEqual("This is implicitly typed plain US-ASCII text." + newline +
                        "It does NOT end with a linebreak.",
                        mime.SubParts[0].ReadContentAsText());
        
        Assert.IsTrue(mime.SubParts[1].Headers.Contains("Content-Type"));
        Assert.AreEqual(MimeType.TextPlain, mime.SubParts[1].MimeType);
        Assert.AreEqual(Charsets.ASCII, mime.SubParts[1].Charset);
        Assert.AreEqual("text/plain; charset=us-ascii", mime.SubParts[1].Headers["Content-Type"].Value);
        Assert.AreEqual("This is explicitly typed plain US-ASCII text." + newline +
                        "It DOES end with a linebreak." + newline,
                        mime.SubParts[1].ReadContentAsText());
      });
    }
    
    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestParseBodyComposite2(string newline, LoadMessageMode mode)
    {
      var content = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", newline);
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.MultipartAlternative, mime.MimeType);
        Assert.AreEqual("==boundary==", mime.Boundary);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("multipart/alternative", mime.Headers["Content-Type"].GetValueWithoutParameter());
        Assert.AreEqual("==boundary==", mime.Headers["Content-Type"].GetParameter("boundary"));
        Assert.IsTrue(mime.Headers.Contains("Subject"));
        Assert.AreEqual("multi part", mime.Headers["Subject"].Value);
        Assert.AreEqual("multipart message body" + newline, mime.ReadContentAsText());
        
        Assert.AreEqual(2, mime.SubParts.Count);
        
        Assert.AreEqual(MimeType.TextPlain, mime.SubParts[0].MimeType);
        Assert.IsTrue(mime.SubParts[0].Headers.Contains("Content-Type"));
        Assert.AreEqual("text/plain", mime.SubParts[0].Headers["Content-Type"].Value);
        Assert.AreEqual("first part, text/plain" + newline, mime.SubParts[0].ReadContentAsText());
        
        Assert.AreEqual(new MimeType("text", "html"), mime.SubParts[1].MimeType);
        Assert.IsTrue(mime.SubParts[1].Headers.Contains("Content-Type"));
        Assert.AreEqual("text/html", mime.SubParts[1].Headers["Content-Type"].Value);
        Assert.AreEqual("<html><body><p>second part, text/html</p></body></html>",
                        mime.SubParts[1].ReadContentAsText());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestParseBodyCompositeWithNoHeader(LoadMessageMode mode)
    {
      var content = @"Return-Path: <>
Date: Fri, 21 Dec 2007 07:04:35 +0900
MIME-Version: 1.0
Content-Type: multipart/report; report-type=delivery-status;
  boundary=""lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx""
Subject: Warning: could not send message for past 4 hours
Auto-Submitted: auto-generated (warning-timeout)

This is a MIME-encapsulated message

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx

    **********************************************
    **      THIS IS A WARNING MESSAGE ONLY      **
    **  YOU DO NOT NEED TO RESEND YOUR MESSAGE  **
    **********************************************

The original message was received at Fri, 21 Dec 2007 02:06:24 +0900
from xxx.xxxxxx.xxx [127.0.0.1]

   ----- Transcript of session follows -----
451 xxx.xxxxxx.xxx: Name server timeout
Warning: message still undelivered after 4 hours
Will keep trying until message is 5 days old
451 xxx.xxxxxx.xxx: Name server timeout

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx--
";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(new MimeType("multipart", "report"), mime.MimeType);
        Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Boundary);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("multipart/report", mime.Headers["Content-Type"].GetValueWithoutParameter());
        Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Headers["Content-Type"].GetParameter("boundary"));
        
        Assert.AreEqual(1, mime.SubParts.Count);
        
        Assert.AreEqual(0, mime.SubParts[0].Headers.Count);
        Assert.IsTrue(mime.SubParts[0].ReadContentAsText().StartsWith("    ***"));
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestParseBodyEncapsulated(LoadMessageMode mode)
    {
      var content = @"Return-Path: <>
Date: Fri, 21 Dec 2007 07:04:35 +0900
MIME-Version: 1.0
Content-Type: multipart/report; report-type=delivery-status;
  boundary=""lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx""
Subject: Warning: could not send message for past 4 hours
Auto-Submitted: auto-generated (warning-timeout)

This is a MIME-encapsulated message

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx
Content-Type: message/delivery-status

Reporting-MTA: dns; xxx.xxxxxx.xxx
Arrival-Date: Fri, 21 Dec 2007 02:06:24 +0900

Final-Recipient: RFC822; root@xxx.xxxxxx.xxx
X-Actual-Recipient: RFC822; xxxxxx@xxx.xxxxxx.xxx
Action: delayed
Status: 4.4.3
Last-Attempt-Date: Fri, 21 Dec 2007 07:04:35 +0900
Will-Retry-Until: Wed, 26 Dec 2007 02:06:24 +0900

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx
Content-Type: message/rfc822

Return-Path: <root@xxx.xxxxxx.xxx>
Date: Fri, 21 Dec 2007 02:02:15 +0900
Content-Type: text/plain; charset=UTF-8
Auto-Submitted: auto-generated
X-Cron-Env: <MAILTO=root,postmaster,webmaster,clamav>
X-Cron-Env: <SHELL=/bin/sh>
X-Cron-Env: <HOME=/root>
X-Cron-Env: <PATH=/usr/bin:/bin>
X-Cron-Env: <LOGNAME=root>
X-Cron-Env: <USER=root>

ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Update failed. Your network may be down or none of the mirrors listed in freshclam.conf is working.

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx--
";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(new MimeType("multipart", "report"), mime.MimeType);
        Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Boundary);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("multipart/report", mime.Headers["Content-Type"].GetValueWithoutParameter());
        Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Headers["Content-Type"].GetParameter("boundary")); 
        
        Assert.AreEqual(2, mime.SubParts.Count);
        
        // part #1
        Assert.AreEqual(new MimeType("message", "delivery-status"), mime.SubParts[0].MimeType);
        Assert.AreEqual("message/delivery-status", mime.SubParts[0].Headers["Content-Type"].Value);
        Assert.IsEmpty(mime.SubParts[0].ReadContentAsText());
        
        Assert.IsNotNull(mime.SubParts[0].EncapsulatedMessage);
        Assert.AreEqual(2, mime.SubParts[0].EncapsulatedMessage.Headers.Count);
        Assert.AreEqual("Reporting-MTA", mime.SubParts[0].EncapsulatedMessage.Headers[0].Name);
        StringAssert.StartsWith("Final-Recipient:", mime.SubParts[0].EncapsulatedMessage.ReadContentAsText());
        
        // part #2
        Assert.AreEqual(new MimeType("message", "rfc822"), mime.SubParts[1].MimeType);
        Assert.AreEqual("message/rfc822", mime.SubParts[1].Headers["Content-Type"].Value);
        Assert.IsEmpty(mime.SubParts[1].ReadContentAsText());
        
        Assert.IsNotNull(mime.SubParts[1].EncapsulatedMessage);
        Assert.AreEqual(10, mime.SubParts[1].EncapsulatedMessage.Headers.Count);
        Assert.AreEqual("Return-Path", mime.SubParts[1].EncapsulatedMessage.Headers[0].Name);
        StringAssert.StartsWith("ERROR: Can't query", mime.SubParts[1].EncapsulatedMessage.ReadContentAsText());
      });
    }
    
    [TestCase("multipart/mixed", LoadMessageMode.OnMemory)]
    [TestCase("multipart/mixed", LoadMessageMode.FileStream)]
    [TestCase("MULTIPART/MIXED", LoadMessageMode.OnMemory)]
    [TestCase("MULTIPART/MIXED", LoadMessageMode.FileStream)]
    public void TestParseBodyMediaTypeCaseInsensitivityBodyComposite(string mediaType, LoadMessageMode mode)
    {
      var content = "Content-Type: " + mediaType + @"; boundary=""boundary""

--boundary
Content-Type: text/plain

part1
--boundary--
";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.IsTrue(MimeType.MultipartMixed.EqualsIgnoreCase(mime.MimeType), "media type");
        
        Assert.AreEqual(1, mime.SubParts.Count, "SubParts.Count");
        Assert.IsNull(mime.EncapsulatedMessage, "EncapsulatedMessage");
      });
    }
    
    [TestCase("message/partial", LoadMessageMode.OnMemory)]
    [TestCase("message/partial", LoadMessageMode.FileStream)]
    [TestCase("MESSAGE/PARTIAL", LoadMessageMode.OnMemory)]
    [TestCase("MESSAGE/PARTIAL", LoadMessageMode.FileStream)]
    public void TestParseBodyMediaTypeCaseInsensitivityBodyPartial(string mediaType, LoadMessageMode mode)
    {
      var content = "Content-Type: " + mediaType + @"; number=1

Content-Type: text/plain

encapsulated message
";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.IsTrue(MimeType.MessagePartial.EqualsIgnoreCase(mime.MimeType), "media type");
        
        Assert.IsNull(mime.EncapsulatedMessage, "EncapsulatedMessage");
        Assert.AreEqual(0, mime.SubParts.Count, "SubParts.Count");
      });
    }
    
    [TestCase("message/rfc822", LoadMessageMode.OnMemory)]
    [TestCase("message/rfc822", LoadMessageMode.FileStream)]
    [TestCase("MESSAGE/RFC822", LoadMessageMode.OnMemory)]
    [TestCase("MESSAGE/RFC822", LoadMessageMode.FileStream)]
    public void TestParseBodyMediaTypeCaseInsensitivityBodyEncapsulated(string mediaType, LoadMessageMode mode)
    {
      var content = "Content-Type: " + mediaType + @"

Content-Type: text/plain

encapsulated message
";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.IsTrue(mime.MimeType.EqualsIgnoreCase(new MimeType("message", "rfc822")), "media type");
        
        Assert.IsNotNull(mime.EncapsulatedMessage, "EncapsulatedMessage");
        Assert.AreEqual(0, mime.SubParts.Count, "SubParts.Count");
      });
    }

    [Test]
    public void TestDecodeHeaderPlainText()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: subject

body");
      
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("subject", mime.Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.None, mime.Headers["Subject"].Encoding);
      Assert.IsNull(mime.Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeader()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=

body");
      
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("件名", mime.Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.BEncoding, mime.Headers["Subject"].Encoding);
      Assert.AreEqual(Charsets.JIS, mime.Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeaderSubPart()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=

first part

--==boundary==--");
      
      Assert.IsTrue(mime.SubParts[0].Headers.Contains("Subject"));
      Assert.AreEqual("件名", mime.SubParts[0].Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.BEncoding, mime.SubParts[0].Headers["Subject"].Encoding);
      Assert.AreEqual(Charsets.JIS, mime.SubParts[0].Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeaderInvalidCharsetSelectFallback()
    {
      bool called = false;
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: =?invalid?B?GyRCN29MPhsoQg==?=

body", delegate(string name) {
        called = true;
        return Charsets.JIS;
      });
      
      Assert.IsTrue(called);
      
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("件名", mime.Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.BEncoding, mime.Headers["Subject"].Encoding);
      Assert.AreEqual(Charsets.JIS, mime.Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeaderInvalidCharsetSelectFallbackReturnNull()
    {
      bool called = false;
      
      try {
        MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: =?invalid?B?GyRCN29MPhsoQg==?=

body", delegate(string name) {
          called = true;
          return null;
        }
        );
        
        Assert.Fail("EncodingNotSupportedException not thrown");
      }
      catch (EncodingNotSupportedException ex) {
        Assert.AreEqual("invalid", ex.EncodingName, "EncodingName");
      }
      
      Assert.IsTrue(called);
    }

    [Test]
    public void TestDecodeHeaderInvalidFormat()
    {
      var mime1 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: =?utf-8?b?****?=

body");

      Assert.AreEqual("=?utf-8?b?****?=", mime1.Headers["Subject"].Value);

      var mime2 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: =?utf-8?q?===?=

body");

      Assert.AreEqual("=?utf-8?q?===?=", mime2.Headers["Subject"].Value);

      var mime3 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: =?utf-8?x?hoge?=

body");

      Assert.AreEqual("=?utf-8?x?hoge?=", mime3.Headers["Subject"].Value);
    }

    [Test]
    public void TestDecodeHeaderInvalidFormatReplaceToAlternativeText()
    {
      var mime1 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: =?utf-8?b?****?=

body", null, (c, m, t) => "foo");

      Assert.AreEqual("foo", mime1.Headers["Subject"].Value);

      var mime2 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: =?utf-8?q?===?=

body", null, (c, m, t) => "bar");

      Assert.AreEqual("bar", mime2.Headers["Subject"].Value);

      var mime3 = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: =?utf-8?x?hoge?=

body", null, (c, m, t) => "baz");

      Assert.AreEqual("baz", mime3.Headers["Subject"].Value);
    }

    [Test]
    public void TestDecodeHeaderInvalidFormatReplaceToAlternativeTextSubPart()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain
Subject: =?utf-8?x?hoge?=

first part

--==boundary==--", null, (c, m, t) => "<alternate>");

      Assert.AreEqual("<alternate>", mime.SubParts[0].Headers["Subject"].Value);
    }
  }
}